from __future__ import absolute_import, division, print_function

import pytest
from operator import (add, sub, mul, floordiv, mod, pow, truediv, eq, ne, lt,
                      gt, le, ge, getitem)

from functools import partial
from datetime import datetime
from datashape.util.testing import assert_dshape_equal
from datashape.predicates import iscollection, isscalar
from blaze.expr import (
    Field,
    by,
    common_subexpression,
    cos,
    count,  # used in eval string
    discover,
    distinct,
    exp,
    join,
    label,  # used in eval string
    isnan,  # used in eval string
    max,
    merge,
    min,
    projection,
    selection,
    sum,
    summary,
    symbol,
    transform,
)
from blaze.compatibility import PY3, builtins
from blaze.utils import raises
from datashape import dshape, int32, int64, Record, DataShape
from toolz import identity


def test_dshape():
    t = symbol('t', 'var * {name: string, amount: int}')
    assert t.dshape == dshape('var * {name: string, amount: int}')


def test_length():
    t = symbol('t', '10 * {name: string, amount: int}')
    s = symbol('s', 'var * {name:string, amount:int}')
    assert t.dshape == dshape('10 * {name: string, amount: int}')
    assert len(t) == 10
    assert len(t.name) == 10
    assert len(t[['name']]) == 10
    assert len(t.sort('name')) == 10
    assert len(t.head(5)) == 5
    assert len(t.head(50)) == 10
    with pytest.raises(ValueError):
        len(s)


def test_symbol_eq():
    assert not (symbol('t', 'var * {name: string}') ==
                symbol('v', 'var * {name: string}'))


def test_symbol_name():
    t = symbol('t', '10 * {people: string, amount: int}')
    r = symbol('r', 'var * int64')
    with pytest.raises(AttributeError):
        t.name
    with pytest.raises(AttributeError):
        r.name


def test_shape():
    t = symbol('t', 'var * {name: string, amount: int}')
    assert t.shape
    assert isinstance(t.shape, tuple)
    assert len(t.shape) == 1


def test_eq():
    assert symbol('t', 'var * {a: string, b: int}').isidentical(
        symbol('t', 'var * {a: string, b: int}'),
    )
    assert not symbol('t', 'var * {b: string, a: int}').isidentical(
        symbol('t', 'var * {a: string, b: int}'),
    )


def test_column():
    t = symbol('t', 'var * {name: string, amount: int}')
    assert t.fields == ['name', 'amount']

    assert eval(str(t.name)) == t.name
    assert str(t.name) == "t.name"
    with pytest.raises(AttributeError):
        t.name.balance
    with pytest.raises((NotImplementedError, ValueError)):
        getitem(t, set('balance'))


def test_symbol_projection_failures():
    t = symbol('t', '10 * {name: string, amount: int}')
    with pytest.raises(ValueError):
        t._project(['name', 'id'])
    with pytest.raises(AttributeError):
        t.foo
    with pytest.raises(TypeError):
        t._project(t.dshape)


def test_Projection():
    t = symbol('t', 'var * {name: string, amount: int, id: int32}')
    p = projection(t, ['amount', 'name'])
    assert p.schema == dshape('{amount: int32, name: string}')
    print(t['amount'].dshape)
    print(dshape('var * int32'))
    assert t['amount'].dshape == dshape('var * int32')
    assert t['amount']._name == 'amount'

    assert eval(str(p)).isidentical(p)
    assert p._project(['amount', 'name']) == p[['amount', 'name']]
    with pytest.raises(ValueError):
        p._project('balance')


def test_Projection_retains_shape():
    t = symbol('t', '5 * {name: string, amount: int, id: int32}')

    assert_dshape_equal(
        t[['name', 'amount']].dshape,
        dshape('5 * {name: string, amount: int}')
    )


def test_indexing():
    t = symbol('t', 'var * {name: string, amount: int, id: int}')
    assert t[['amount', 'id']] == projection(t, ['amount', 'id'])
    assert t['amount'].isidentical(Field(t, 'amount'))


def test_relational():
    t = symbol('t', 'var * {name: string, amount: int, id: int}')

    r = (t['name'] == 'Alice')

    assert 'bool' in str(r.dshape)
    assert r._name


def test_selection():
    t = symbol('t', 'var * {name: string, amount: int, id: int}')

    s = selection(t, t['name'] == 'Alice')
    f = selection(t, t['id'] > t['amount'])
    p = t[t['amount'] > 100]
    with pytest.raises(ValueError):
        selection(t, p)

    assert s.dshape == t.dshape


def test_selection_typecheck():
    t = symbol('t', 'var * {name: string, amount: int, id: int}')

    assert raises(TypeError, lambda: t[t['amount'] + t['id']])
    assert raises(TypeError, lambda: t[t['name']])


def test_selection_by_indexing():
    t = symbol('t', 'var * {name: string, amount: int, id: int}')

    result = t[t['name'] == 'Alice']

    assert t.schema == result.schema
    assert 'Alice' in str(result)


def test_selection_by_getattr():
    t = symbol('t', 'var * {name: string, amount: int, id: int}')

    result = t[t.name == 'Alice']

    assert t.schema == result.schema
    assert 'Alice' in str(result)


def test_selection_path_check():
    t = symbol('t', 'var * {name: string, amount: int, id: int}')
    t2 = t[t.name == 'Alice']
    t3 = t2[t2.amount > 0]


def test_path_issue():
    t = symbol('t', "{topic: string, word: string, result: ?float64}")
    t2 = transform(t, sizes=t.result.map(lambda x: (x - MIN)*10/(MAX - MIN),
                                         schema='float64', name='size'))

    assert builtins.any(t2.sizes.isidentical(node) for node in t2.children)


def test_getattr_doesnt_override_properties():
    t = symbol('t', 'var * {_subs: string, schema: string}')
    assert callable(t._subs)
    assert isinstance(t.schema, DataShape)


def test_dir_contains_columns():
    t = symbol('t', 'var * {name: string, amount: int, id: int}')
    result = dir(t)
    columns_set = set(t.fields)
    assert set(result) & columns_set == columns_set


def test_selection_consistent_children():
    t = symbol('t', 'var * {name: string, amount: int, id: int}')

    expr = t['name'][t['amount'] < 0]

    assert list(expr.fields) == ['name']


def test_str():
    import re
    t = symbol('t', 'var * {name: string, amount: int, id: int}')
    expr = t[t['amount'] < 0]['id'] * 2
    assert '<class' not in str(expr)
    assert not re.search('0x[0-9a-f]+', str(expr))

    assert eval(str(expr)) == expr

    assert '*' in str(expr)


def test_join():
    t = symbol('t', 'var * {name: string, amount: int}')
    s = symbol('t', 'var * {name: string, id: int}')
    r = symbol('r', 'var * {name: string, amount: int}')
    q = symbol('q', 'var * {name: int}')

    j = join(t, s, 'name', 'name')

    assert j.schema == dshape('{name: string, amount: int, id: int}')

    assert join(t, s, 'name') == join(t, s, 'name')

    assert join(t, s, 'name').on_left == 'name'
    assert join(t, s, 'name').on_right == 'name'

    assert join(t, r, ('name', 'amount')).on_left == ['name', 'amount']
    with pytest.raises(TypeError):
        join(t, q, 'name')
    with pytest.raises(ValueError):
        join(t, s, how='upside_down')


def test_join_different_on_right_left_columns():
    t = symbol('t', 'var * {x: int, y: int}')
    s = symbol('t', 'var * {a: int, b: int}')
    j = join(t, s, 'x', 'a')
    assert j.on_left == 'x'
    assert j.on_right == 'a'


def test_joined_column_first_in_schema():
    t = symbol('t', 'var * {x: int, y: int, z: int}')
    s = symbol('s', 'var * {w: int, y: int}')

    assert join(t, s).schema == dshape('{y: int, x: int, z: int, w: int}')


def test_outer_join():
    t = symbol('t', 'var * {name: string, amount: int}')
    s = symbol('t', 'var * {name: string, id: int}')

    jleft = join(t, s, 'name', 'name', how='left')
    jright = join(t, s, 'name', 'name', how='right')
    jinner = join(t, s, 'name', 'name', how='inner')
    jouter = join(t, s, 'name', 'name', how='outer')

    js = [jleft, jright, jinner, jouter]

    assert len(set(js)) == 4  # not equal

    assert jinner.schema == dshape('{name: string, amount: int, id: int}')
    assert jleft.schema == dshape('{name: string, amount: int, id: ?int}')
    assert jright.schema == dshape('{name: string, amount: ?int, id: int}')
    assert jouter.schema == dshape('{name: string, amount: ?int, id: ?int}')

    # Default behavior
    assert (join(t, s, 'name', 'name', how='inner') ==
            join(t, s, 'name', 'name'))


def test_join_default_shared_columns():
    t = symbol('t', 'var * {name: string, amount: int}')
    s = symbol('t', 'var * {name: string, id: int}')
    assert join(t, s) == join(t, s, 'name', 'name')


def test_multi_column_join():
    a = symbol('a', 'var * {x: int, y: int, z: int}')
    b = symbol('b', 'var * {w: int, x: int, y: int}')
    j = join(a, b, ['x', 'y'])

    assert set(j.fields) == set('wxyz')

    assert j.on_left == j.on_right == ['x', 'y']
    assert hash(j)

    assert j.fields == ['x', 'y', 'z', 'w']


def test_traverse():
    t = symbol('t', 'var * {name: string, amount: int}')
    assert t in list(t._traverse())

    expr = t.amount.sum()
    trav = list(expr._traverse())
    assert builtins.any(t.amount.isidentical(x) for x in trav)


def test_unary_ops():
    t = symbol('t', 'var * {name: string, amount: int}')
    expr = cos(exp(t['amount']))
    assert 'cos' in str(expr)

    assert '~' in str(~(t.amount > 0))


def test_reduction():
    t = symbol('t', 'var * {name: string, amount: int32}')
    r = sum(t['amount'])
    assert r.dshape in (dshape('int64'),
                        dshape('{amount: int64}'),
                        dshape('{amount_sum: int64}'))

    assert 'amount' not in str(t.count().dshape)

    assert t.count().dshape[0] in (int32, int64)

    assert 'int' in str(t.count().dshape)
    assert 'int' in str(t.nunique().dshape)
    assert 'string' in str(t['name'].max().dshape)
    assert 'string' in str(t['name'].min().dshape)
    assert 'string' not in str(t.count().dshape)

    t = symbol('t', 'var * {name: string, amount: real, id: int}')

    assert 'int' in str(t['id'].sum().dshape)
    assert 'int' not in str(t['amount'].sum().dshape)


def test_reduction_name():
    t = symbol('t', 'var * {name: string, amount: int32, id: int32}')
    assert (t.amount + t.id).sum()._name


def test_max_min_class():
    t = symbol('t', 'var * {name: string, amount: int32}')
    assert str(max(t).dshape) == '{name: string, amount: int32}'
    assert str(min(t).dshape) == '{name: string, amount: int32}'


@pytest.fixture
def symsum():
    t = symbol('t', 'var * {name: string, amount: int32}')
    return t, t.amount.sum()


@pytest.fixture
def ds():
    return dshape("var * {"
                  "transaction_key : int64,"
                  "user_from_key : int64,"
                  "user_to_key : int64,"
                  "date : int64,"
                  "value : float64"
                  "}")


def test_discover_dshape_symbol(ds):
    t_ds = symbol('t', dshape=ds)
    assert t_ds.fields is not None


class TestScalarArithmetic(object):
    ops = {'+': add, '-': sub, '*': mul, '/': truediv, '//': floordiv, '%': mod,
           '**': pow, '==': eq, '!=': ne, '<': lt, '>': gt, '<=': le, '>=': ge}

    def test_scalar_arith(self, symsum):
        def runner(f):
            result = f(r, 1)
            assert eval('r %s 1' % op).isidentical(result)

            a = f(r, r)
            b = eval('r %s r' % op)
            assert a is b or a.isidentical(b)

            result = f(1, r)
            assert eval('1 %s r' % op).isidentical(result)

        t, r = symsum
        r = t.amount.sum()
        for op, f in self.ops.items():
            runner(f)

    def test_scalar_usub(self, symsum):
        t, r = symsum
        result = -r
        assert eval(str(result)).isidentical(result)

    @pytest.mark.xfail
    def test_scalar_uadd(self, symsum):
        t, r = symsum
        +r


def test_summary():
    t = symbol('t', 'var * {id: int32, name: string, amount: int32}')
    s = summary(total=t.amount.sum(), num=t.id.count())
    assert s.dshape == dshape('{num: int32, total: int64}')
    assert hash(s)
    assert eval(str(s)).isidentical(s)

    assert 'summary(' in str(s)
    assert 'total=' in str(s)
    assert 'num=' in str(s)
    assert str(t.amount.sum()) in str(s)

    assert not summary(total=t.amount.sum())._child.isidentical(t.amount.sum())
    assert iscollection(summary(total=t.amount.sum() + 1)._child.dshape)


def test_reduction_arithmetic():
    t = symbol('t', 'var * {id: int32, name: string, amount: int32}')
    expr = t.amount.sum() + 1
    assert eval(str(expr)).isidentical(expr)


def test_Distinct():
    t = symbol('t', 'var * {name: string, amount: int32}')
    r = distinct(t['name'])
    print(r.dshape)
    assert r.dshape == dshape('var * string')
    assert r._name == 'name'

    r = t.distinct()
    assert r.dshape == t.dshape


def test_by():
    t = symbol('t', 'var * {name: string, amount: int32, id: int32}')
    r = by(t['name'], total=sum(t['amount']))

    print(r.schema)
    assert isinstance(r.schema[0], Record)
    assert str(r.schema[0]['name']) == 'string'


def test_by_summary():
    t = symbol('t', 'var * {name: string, amount: int32, id: int32}')
    a = by(t['name'], sum=sum(t['amount']))
    b = by(t['name'], summary(sum=sum(t['amount'])))

    assert a.isidentical(b)


def test_by_summary_printing():
    t = symbol('t', 'var * {name: string, amount: int32, id: int32}')
    assert (str(by(t.name, total=sum(t.amount))) ==
            'by(t.name, total=sum(t.amount))')


def test_by_columns():
    t = symbol('t', 'var * {name: string, amount: int32, id: int32}')

    assert len(by(t['id'], total=t['amount'].sum()).fields) == 2
    assert len(by(t['id'], count=t['id'].count()).fields) == 2
    print(by(t, count=t.count()).fields)
    assert len(by(t, count=t.count()).fields) == 4


def test_sort():
    t = symbol('t', 'var * {name: string, amount: int32, id: int32}')
    s = t.sort('amount', ascending=True)
    print(str(s))
    assert eval(str(s)).isidentical(s)

    assert s.schema == t.schema

    assert t['amount'].sort().key == 'amount'


def test_head():
    t = symbol('t', 'var * {name: string, amount: int32, id: int32}')
    s = t.head(10)
    assert eval(str(s)).isidentical(s)

    assert s.schema == t.schema


def test_label():
    t = symbol('t', 'var * {name: string, amount: int32, id: int32}')
    quantity = (t['amount'] + 100).label('quantity')

    assert eval(str(quantity)).isidentical(quantity)

    assert quantity.fields == ['quantity']

    with pytest.raises(ValueError):
        quantity['balance']


def test_map_label():
    t = symbol('t', 'var * {name: string, amount: int32, id: int32}')
    c = t.amount.map(identity, schema='int32')
    assert c.label('bar')._name == 'bar'
    assert c.label('bar')._child.isidentical(c._child)


def test_columns():
    t = symbol('t', 'var * {name: string, amount: int32, id: int32}')
    assert list(t.fields) == ['name', 'amount', 'id']
    assert list(t['name'].fields) == ['name']
    (t['amount'] + 1).fields


def test_relabel():
    t = symbol('t', 'var * {name: string, amount: int32, id: int32}')

    rl = t.relabel({'name': 'NAME', 'id': 'ID'})
    rlc = t['amount'].relabel({'amount': 'BALANCE'})

    assert eval(str(rl)).isidentical(rl)

    print(rl.fields)
    assert rl.fields == ['NAME', 'amount', 'ID']

    assert not isscalar(rl.dshape.measure)
    assert isscalar(rlc.dshape.measure)


def test_relabel_join():
    names = symbol('names', 'var * {first: string, last: string}')

    siblings = join(names.relabel({'last': 'left'}),
                    names.relabel({'last': 'right'}), 'first')

    assert siblings.fields == ['first', 'left', 'right']


def test_map():
    t = symbol('t', 'var * {name: string, amount: int32, id: int32}')
    inc = lambda x: x + 1
    assert isscalar(t['amount'].map(inc, schema='int').dshape.measure)
    s = t['amount'].map(inc, schema='{amount: int}')
    assert not isscalar(s.dshape.measure)

    assert s.dshape == dshape('var * {amount: int}')

    expr = (t[['name', 'amount']]
            .map(identity, schema='{name: string, amount: int}'))
    assert expr._name is None


@pytest.mark.xfail(reason="Not sure that we should even support this")
def test_map_without_any_info():
    t = symbol('t', 'var * {name: string, amount: int32, id: int32}')
    assert iscolumn(t['amount'].map(inc, 'int'))
    assert not iscolumn(t[['name', 'amount']].map(identity))


def test_apply():
    t = symbol('t', 'var * {name: string, amount: int32, id: int32}')
    s = t['amount'].apply(sum, dshape='real')
    r = t['amount'].apply(sum, dshape='3 * real')
    assert s.dshape == dshape('real')
    assert r.schema == dshape('real')


def test_symbol_printing_is_legible():
    accounts = symbol('accounts',
                      'var * {name: string, balance: int, id: int}')

    expr = (exp(accounts.balance * 10)) + accounts['id']
    assert "exp(accounts.balance * 10)" in str(expr)
    assert "+ accounts.id" in str(expr)


def test_merge():
    t = symbol('t', 'int64')
    p = symbol('p', 'var * {amount:int}')
    accounts = symbol('accounts',
                      'var * {name: string, balance: int32, id: int32}')
    new_amount = (accounts.balance * 1.5).label('new')

    c = merge(accounts[['name', 'balance']], new_amount)
    assert c.fields == ['name', 'balance', 'new']
    assert c.schema == dshape('{name: string, balance: int32, new: float64}')

    with pytest.raises(ValueError):
        merge(t, t)
    with pytest.raises(ValueError):
        merge(t, p)


def test_merge_repeats():
    accounts = symbol('accounts',
                      'var * {name: string, balance: int32, id: int32}')
    with pytest.raises(ValueError):
        merge(accounts, (accounts.balance + 1).label('balance'))


def test_merge_project():
    accounts = symbol('accounts',
                      'var * {name: string, balance: int32, id: int32}')
    new_amount = (accounts['balance'] * 1.5).label('new')
    c = merge(accounts[['name', 'balance']], new_amount)

    assert c['new'].isidentical(new_amount)
    assert c['name'].isidentical(accounts['name'])

    assert c[['name', 'new']].isidentical(merge(accounts.name, new_amount))


inc = lambda x: x + 1


def test_subterms():
    a = symbol('a', 'var * {x: int, y: int, z: int}')
    assert list(a._subterms()) == [a]
    assert set(a['x']._subterms()) == set([a, a['x']])
    assert (set(a['x'].map(inc, 'int')._subterms()) ==
            set([a, a['x'], a['x'].map(inc, 'int')]))
    assert a in set((a['x'] + 1)._subterms())


def test_common_subexpression():
    a = symbol('a', 'var * {x: int, y: int, z: int}')

    assert common_subexpression(a).isidentical(a)
    assert common_subexpression(a, a['x']).isidentical(a)
    assert common_subexpression(a['y'] + 1, a['x']).isidentical(a)
    assert common_subexpression(a['x'].map(inc, 'int'), a['x']).isidentical(
        a['x'],
    )


def test_schema_of_complex_interaction():
    a = symbol('a', 'var * {x: int, y: int, z: int}')
    expr = (a['x'] + a['y']) / a['z']
    assert expr.schema == dshape('float64')

    expr = expr.label('foo')
    assert expr.schema == dshape('float64')


def iscolumn(x):
    return isscalar(x.dshape.measure)


def test_iscolumn():
    a = symbol('a', 'var * {x: int, y: int, z: int}')
    assert not iscolumn(a)
    assert iscolumn(a['x'])
    assert not iscolumn(a[['x', 'y']])
    assert not iscolumn(a[['x']])
    assert iscolumn((a['x'] + a['y']))
    assert iscolumn(a['x'].distinct())
    assert not iscolumn(a[['x']].distinct())
    assert not iscolumn(by(a['x'], total=a['y'].sum()))
    assert iscolumn(a['x'][a['x'] > 1])
    assert not iscolumn(a[['x', 'y']][a['x'] > 1])
    assert iscolumn(a['x'].sort())
    assert not iscolumn(a[['x', 'y']].sort())
    assert iscolumn(a['x'].head())
    assert not iscolumn(a[['x', 'y']].head())

    assert iscolumn(symbol('b', 'int'))
    assert not iscolumn(symbol('b', 'var * {x: int}'))


def test_discover():
    ds = 'var * {x: int, y: int, z: int}'
    a = symbol('a', ds)
    assert discover(a) == dshape(ds)


def test_improper_selection():
    t = symbol('t', 'var * {x: int, y: int, z: int}')

    assert raises(Exception, lambda: t[t['x'] > 0][t.sort()[t['y' > 0]]])


def test_serializable():
    t = symbol('t', 'var * {id: int, name: string, amount: int}')
    import pickle
    t2 = pickle.loads(pickle.dumps(t, protocol=pickle.HIGHEST_PROTOCOL))

    assert t.isidentical(t2)

    s = symbol('t', 'var * {id: int, city: string}')
    expr = join(t[t.amount < 0], s).sort('id').city.head()
    expr2 = pickle.loads(pickle.dumps(expr, protocol=pickle.HIGHEST_PROTOCOL))

    assert expr.isidentical(expr2)


def test_symbol_coercion():
    from datetime import date
    t = symbol('t', 'var * {name: string, amount: int, timestamp: ?date}')
    assert (t.amount + '10').rhs == 10

    assert (t.timestamp < '2014-12-01').rhs == date(2014, 12, 1)


def test_isnan():
    t = symbol('t', 'var * {name: string, amount: real, timestamp: ?date}')

    for expr in [t.amount.isnan(), ~t.amount.isnan()]:
        assert eval(str(expr)).isidentical(expr)

    assert iscollection(t.amount.isnan().dshape)
    assert 'bool' in str(t.amount.isnan().dshape)


def test_distinct_name():
    t = symbol('t', 'var * {id: int32, name: string}')

    assert t.name.isidentical(t['name'])
    assert t.distinct().name.isidentical(t.distinct()['name'])
    assert t.id.distinct()._name == 'id'
    assert t.name._name == 'name'


def test_leaves():
    t = symbol('t', 'var * {id: int32, name: string}')
    v = symbol('v', 'var * {id: int32, city: string}')
    x = symbol('x', 'int32')

    assert t._leaves() == [t]
    assert t.id._leaves() == [t]
    assert by(t.name, count=t.id.nunique())._leaves() == [t]
    assert join(t, v)._leaves() == [t, v]
    assert join(v, t)._leaves() == [v, t]

    assert (x + 1)._leaves() == [x]


@pytest.fixture
def t():
    return symbol('t', 'var * {id: int, amount: float64, name: string}')


def funcname(x, y='<lambda>'):
    if PY3:
        return 'TestRepr.%s.<locals>.%s' % (x, y)
    return 'test_symbol.%s' % y


class TestRepr(object):
    def test_partial_lambda(self, t):
        expr = t.amount.map(partial(lambda x, y: x + y, 1))
        s = str(expr)
        assert s == ("Map(_child=t.amount, "
                     "func=partial(%s, 1), "
                     "_asschema=None, _name0=None)" %
                     funcname('test_partial_lambda'))

    def test_lambda(self, t):
        expr = t.amount.map(lambda x: x)
        s = str(expr)
        assert s == ("Map(_child=t.amount, "
                     "func=%s, _asschema=None, _name0=None)" %
                     funcname('test_lambda'))

    def test_partial(self, t):
        def myfunc(x, y):
            return x + y
        expr = t.amount.map(partial(myfunc, 1))
        s = str(expr)
        assert s == ("Map(_child=t.amount, "
                     "func=partial(%s, 1), "
                     "_asschema=None, _name0=None)" % funcname('test_partial',
                                                               'myfunc'))

    def test_builtin(self, t):
        expr = t.amount.map(datetime.fromtimestamp)
        s = str(expr)
        assert s == ("Map(_child=t.amount, "
                     "func=datetime.fromtimestamp, _asschema=None,"
                     " _name0=None)")

    def test_udf(self, t):
        def myfunc(x):
            return x + 1
        expr = t.amount.map(myfunc)
        s = str(expr)
        assert s == ("Map(_child=t.amount, "
                     "func=%s, _asschema=None,"
                     " _name0=None)" % funcname('test_udf', 'myfunc'))

    def test_nested_partial(self, t):
        def myfunc(x, y, z):
            return x + y + z
        f = partial(partial(myfunc, 2), 1)
        expr = t.amount.map(f)
        s = str(expr)
        fn = funcname('test_nested_partial', 'myfunc')
        assert s in ("Map(_child=t.amount, func=partial(partial(%s, 2), 1),"
                     " _asschema=None, _name0=None)" % fn,
                     # Py3.5 version does partial() folding!
                     "Map(_child=t.amount, func=partial(%s, 2, 1),"
                     " _asschema=None, _name0=None)" % fn)


def test_count_values():
    t = symbol('t', 'var * {name: string, amount: int, city: string}')
    assert t.name.count_values(sort=False).isidentical(
        by(t.name, count=t.name.count()),
    )
    assert t.name.count_values(sort=True).isidentical(
        by(t.name, count=t.name.count()).sort('count', ascending=False),
    )


def test_dir():
    t = symbol('t', 'var * {name: string, amount: int, dt: datetime}')
    assert 'day' in dir(t.dt)
    assert 'mean' not in dir(t.dt)
    assert 'mean' in dir(t.amount)
    assert 'like' not in dir(t[['amount', 'dt']])
    assert 'any' not in dir(t.name)


def test_distinct_column():
    t = symbol('t', 'var * {name: string, amount: int, dt: datetime}')
    assert t.name.distinct().name.dshape == t.name.distinct().dshape
    assert t.name.distinct().name.isidentical(t.name.distinct())



from blaze.compute.csv import pre_compute, CSV
from blaze import compute, discover, dshape, into, join, concat, data
from blaze.utils import example, filetext, filetexts
from blaze.expr import symbol
from pandas import DataFrame, Series
import pandas.util.testing as tm
from datashape.predicates import iscollection
import numpy as np
import pandas as pd
from toolz import first
from collections import Iterator
from odo import odo
from odo.chunks import chunks
import dask.dataframe

def test_pre_compute_on_small_csv_gives_dataframe():
    csv = CSV(example('iris.csv'))
    s = symbol('s', discover(csv))
    assert isinstance(pre_compute(s.species, csv), (Series, DataFrame))


def test_pre_compute_on_large_csv_gives_dask_reader():
    csv = CSV(example('iris.csv'))
    s = symbol('s', discover(csv))
    assert isinstance(pre_compute(s.species, csv, comfortable_memory=10),
                      dask.dataframe.DataFrame)


def test_pre_compute_with_head_on_large_csv_yields_iterator():
    csv = CSV(example('iris.csv'))
    s = symbol('s', discover(csv))
    assert isinstance(pre_compute(s.species.head(), csv, comfortable_memory=10),
                      Iterator)


def test_compute_chunks_on_single_csv():
    csv = CSV(example('iris.csv'))
    s = symbol('s', discover(csv))
    expr = s.sepal_length.max()
    assert compute(expr, {s: csv}, comfortable_memory=10, blocksize=50) == 7.9


def test_compute_with_projection_projects_on_data_frames():
    csv = CSV(example('iris.csv'))
    s = symbol('s', discover(csv))
    result = compute(s[['sepal_length', 'sepal_width']],
                     csv, comfortable_memory=10)
    assert set(result.columns) == set(['sepal_length', 'sepal_width'])


def test_compute_calls_lean_projection():
    csv = CSV(example('iris.csv'))
    s = symbol('s', discover(csv))
    result = pre_compute(s[s.sepal_length > 5.0].species,
                         csv, comfortable_memory=10)
    assert set(result.columns) == set(['sepal_length', 'species'])


def test_unused_datetime_columns():
    ds = dshape('2 * {val: string, when: datetime}')
    with filetext("val,when\na,2000-01-01\nb,2000-02-02") as fn:
        csv = CSV(fn, has_header=True)

        s = symbol('s', discover(csv))
        assert into(list, compute(s.val, csv)) == ['a', 'b']


def test_multiple_csv_files():
    d = {'mult1.csv': 'name,val\nAlice,1\nBob,2',
         'mult2.csv': 'name,val\nAlice,3\nCharlie,4'}

    dta = [('Alice', 1), ('Bob', 2), ('Alice', 3), ('Charlie', 4)]
    with filetexts(d) as fns:
        r = data('mult*.csv')
        s = symbol('s', discover(r))

        for e in [s, s.name, s.name.nunique(), s.name.count_values(),
                s.val.mean()]:
            a = compute(e, {s: r})
            b = compute(e, {s: dta})
            if iscollection(e.dshape):
                a, b = into(set, a), into(set, b)
            assert a == b


def test_csv_join():
    d = {'a.csv': 'a,b,c\n0,1,2\n3,4,5',
         'b.csv': 'c,d,e\n2,3,4\n5,6,7'}

    with filetexts(d):
        data_a = data('a.csv')
        data_b = data('b.csv')
        a = symbol('a', discover(data_a))
        b = symbol('b', discover(data_b))
        tm.assert_frame_equal(
            odo(
                compute(join(a, b, 'c'), {a: data_a, b: data_b}),
                pd.DataFrame,
            ),

            # windows needs explicit int64 construction b/c default is int32
            pd.DataFrame(np.array([[2, 0, 1, 3, 4],
                                   [5, 3, 4, 6, 7]], dtype='int64'),
                         columns=list('cabde'))
        )


def test_concat():
    d = {'a.csv': 'a,b\n1,2\n3,4',
         'b.csv': 'a,b\n5,6\n7,8'}

    with filetexts(d):
        a_rsc = data('a.csv')
        b_rsc = data('b.csv')

        a = symbol('a', discover(a_rsc))
        b = symbol('b', discover(b_rsc))

        tm.assert_frame_equal(
            odo(
                compute(concat(a, b), {a: a_rsc, b: b_rsc}), pd.DataFrame,
            ),

            # windows needs explicit int64 construction b/c default is int32
            pd.DataFrame(np.arange(1, 9, dtype='int64').reshape(4, 2),
                         columns=list('ab')),
        )



from __future__ import absolute_import, division, print_function

import sqlalchemy as sa

from .compatibility import basestring
from .dispatch import dispatch

__all__ = ()


@dispatch(sa.Table, basestring)
def create_index(s, column, **kwargs):
    """Create an index for a single column.

    Parameters
    ----------
    s : sa.Table
        The table to create the index on.
    column : str
        The name of the column to create an index on.
    name : str
        The name of the created index.
    unique : bool, optional
        Should the index be unique.
    ignore_existing : bool, optional
        Should this supress exceptions from the index already existing.
    concurrently : bool, optional
        Should the index be created without holding a lock. This feature is
        postgres specific.
    """
    return create_index(s, (column,), **kwargs)


@dispatch(sa.Table, (list, tuple))
def create_index(s,
                 columns,
                 name=None,
                 unique=False,
                 ignore_existing=False,
                 concurrently=False):
    """Create an index for a single column.

    Parameters
    ----------
    s : sa.Table
        The table to create the index on.
    columns : list[str] or tuple[str]
        The names of the columns to create an index on.
    name : str
        The name of the created index.
    unique : bool, optional
        Should the index be unique.
    ignore_existing : bool, optional
        Should this supress exceptions from the index already existing.
    concurrently : bool, optional
        Should the index be created without holding a lock. This feature is
        postgres specific.
    """

    if name is None:
        raise ValueError('SQL indexes must have a name')
    try:
        sa.Index(
            name,
            *(s.c[c] for c in columns),
            unique=unique,
            **{'postgresql_concurrently': concurrently} if concurrently else {}
        ).create(s.bind)
    except (sa.exc.ProgrammingError, sa.exc.OperationalError):
        # sqlite raises OperationalError but postgres raises ProgrammingError
        # thanks fam
        if not ignore_existing:
            raise



import textwrap
import datetime
import pickle
import sys
from types import MethodType

import dask.array as da
from datashape import dshape
from datashape.util.testing import assert_dshape_equal
import pandas as pd
import pandas.util.testing as tm
import pytest
import numpy as np
from odo import into, append
from odo.backends.csv import CSV

from blaze import discover, transform
from blaze.compatibility import pickle
from blaze.expr import symbol
from blaze.interactive import (
    coerce_core,
    compute,
    concrete_head,
    data,
    expr_repr,
    iscorescalar,
    iscoresequence,
    iscoretype,
    to_html,
)
from blaze.utils import tmpfile, example

tdata = (('Alice', 100),
         ('Bob', 200))

L = [[1, 'Alice',   100],
     [2, 'Bob',    -200],
     [3, 'Charlie', 300],
     [4, 'Denis',   400],
     [5, 'Edith',  -500]]

t = data(tdata, fields=['name', 'amount'])

x = np.ones((2, 2))


def test_discover_on_data():
    assert discover(t) == dshape("2 * {name: string, amount: int64}")


def test_table_raises_on_inconsistent_inputs():
    with pytest.raises(ValueError) as excinfo:
        data(tdata, schema='{name: string, amount: float32}',
             dshape=dshape("{name: string, amount: float32}"))
    assert "specify one of schema= or dshape= keyword" in str(excinfo.value)


def test_resources():
    assert t._resources() == {t: t.data}


def test_resources_fail():
    t = symbol('t', 'var * {x: int, y: int}')
    d = t[t['x'] > 100]
    with pytest.raises(ValueError):
        compute(d)


def test_compute_on_Data_gives_back_data():
    assert compute(data([1, 2, 3])) == [1, 2, 3]


def test_len():
    assert len(t) == 2
    assert len(t.name) == 2


def test_compute():
    assert list(compute(t['amount'] + 1)) == [101, 201]


def test_create_with_schema():
    t = data(tdata, schema='{name: string, amount: float32}')
    assert t.schema == dshape('{name: string, amount: float32}')


def test_create_with_raw_data():
    t = data(tdata, fields=['name', 'amount'])
    assert t.schema == dshape('{name: string, amount: int64}')
    assert t.name
    assert t.data == tdata


def test_repr():
    result = expr_repr(t['name'])
    print(result)
    assert isinstance(result, str)
    assert 'Alice' in result
    assert 'Bob' in result
    assert '...' not in result

    result = expr_repr(t['amount'] + 1)
    print(result)
    assert '101' in result

    t2 = data(tuple((i, i**2) for i in range(100)), fields=['x', 'y'])
    assert t2.dshape == dshape('100 * {x: int64, y: int64}')

    result = expr_repr(t2)
    print(result)
    assert len(result.split('\n')) < 20
    assert '...' in result


def test_str_does_not_repr():
    # see GH issue #1240.
    d = data([('aa', 1), ('b', 2)], name="ZZZ",
             dshape='2 * {a: string, b: int64}')
    expr = transform(d, c=d.a.str.len() + d.b)
    assert (str(expr) ==
            "Merge(_child=ZZZ, children=(ZZZ, label(len(_child=ZZZ.a)"
            " + ZZZ.b, 'c')))")


def test_repr_of_scalar():
    assert expr_repr(t.amount.sum()) == '300'


def test_mutable_backed_repr():
    mutable_backed_table = data([[0]], fields=['col1'])
    expr_repr(mutable_backed_table)


def test_dataframe_backed_repr():
    df = pd.DataFrame(data=[0], columns=['col1'])
    dataframe_backed_table = data(df)
    expr_repr(dataframe_backed_table)


def test_dataframe_backed_repr_complex():
    df = pd.DataFrame([(1, 'Alice', 100),
                       (2, 'Bob', -200),
                       (3, 'Charlie', 300),
                       (4, 'Denis', 400),
                       (5, 'Edith', -500)],
                      columns=['id', 'name', 'balance'])
    t = data(df)
    expr_repr(t[t['balance'] < 0])


def test_repr_html_on_no_resources_symbol():
    t = symbol('t', '5 * {id: int, name: string, balance: int}')
    assert to_html(t) == 't'


def test_expr_repr_empty():
    s = expr_repr(t[t.amount > 1e9])
    assert isinstance(s, str)
    assert 'amount' in s


def test_to_html():
    s = to_html(t)
    assert s
    assert 'Alice' in s
    assert '<table' in s

    assert to_html(1) == '1'

    assert to_html(t.count()) == '2'


def test_to_html_on_arrays():
    s = to_html(data(np.ones((2, 2))))
    assert '1' in s
    assert 'br>' in s


def test_repr_html():
    assert '<table' in t._repr_html_()
    assert '<table' in t.name._repr_html_()


def test_into():
    assert into(list, t) == into(list, tdata)


def test_serialization():
    t2 = pickle.loads(pickle.dumps(t, protocol=pickle.HIGHEST_PROTOCOL))

    assert t.schema == t2.schema
    assert t._name == t2._name


def test_table_resource():
    with tmpfile('csv') as filename:
        ds = dshape('var * {a: int, b: int}')
        csv = CSV(filename)
        append(csv, [[1, 2], [10, 20]], dshape=ds)

        t = data(filename)
        assert isinstance(t.data, CSV)
        assert into(list, compute(t)) == into(list, csv)


def test_explicit_override_dshape():
    ds = dshape("""var * {a: ?float64,
                        b: ?string,
                        c: ?float32}""")
    # If not overridden, the dshape discovery will return:
    # var * {a: int64, b: string, c: int64}.
    s = textwrap.dedent("""\
                        a,b,c
                        1,x,3
                        2,y,4
                        3,z,5
                        """)
    with tmpfile('.csv') as filename:
        with open(filename, 'w') as fd:
            fd.write(s)
        bdf = data(filename, dshape=ds)
        assert bdf.dshape == ds


def test_concretehead_failure():
    t = symbol('t', 'var * {x:int, y:int}')
    d = t[t['x'] > 100]
    with pytest.raises(ValueError):
        concrete_head(d)


def test_into_np_ndarray_column():
    t = data(L, fields=['id', 'name', 'balance'])
    expr = t[t.balance < 0].name
    colarray = into(np.ndarray, expr)
    assert len(list(compute(expr))) == len(colarray)


def test_into_nd_array_selection():
    t = data(L, fields=['id', 'name', 'balance'])
    expr = t[t['balance'] < 0]
    selarray = into(np.ndarray, expr)
    assert len(list(compute(expr))) == len(selarray)


def test_into_nd_array_column_failure():
    tble = data(L, fields=['id', 'name', 'balance'])
    expr = tble[tble['balance'] < 0]
    colarray = into(np.ndarray, expr)
    assert len(list(compute(expr))) == len(colarray)


def test_Data_attribute_repr():
    t = data(CSV(example('accounts-datetimes.csv')))
    result = t.when.day
    expected = pd.DataFrame({'when_day': [1, 2, 3, 4, 5]})
    assert expr_repr(result) == repr(expected)


def test_can_trivially_create_csv_data():
    data(example('iris.csv'))

    # in context
    with data(example('iris.csv')) as d:
        assert d is not None


def test_can_trivially_create_csv_Data_with_unicode():
    if sys.version[0] == '2':
        assert isinstance(data(example(u'iris.csv')).data, CSV)


def test_can_trivially_create_sqlite_table():
    pytest.importorskip('sqlalchemy')
    data('sqlite:///'+example('iris.db')+'::iris')

    # in context
    with data('sqlite:///'+example('iris.db')+'::iris') as d:
        assert d is not None


@pytest.mark.xfail(sys.platform != 'darwin', reason="h5py/pytables mismatch")
@pytest.mark.skipif(sys.version_info[:2] == (3, 4) and sys.platform == 'win32',
                    reason='PyTables + Windows + Python 3.4 crashes')
def test_can_trivially_create_pytables():
    pytest.importorskip('tables')
    with data(example('accounts.h5')+'::/accounts') as d:
        assert d is not None


def test_data_passes_kwargs_to_resource():
    assert data(example('iris.csv'), encoding='ascii').data.encoding == 'ascii'


def test_data_on_iterator_refies_data():
    tdata = [1, 2, 3]
    d = data(iter(tdata))

    assert into(list, d) == tdata
    assert into(list, d) == tdata

    # in context
    with data(iter(tdata)) as d:
        assert d is not None


def test_Data_on_json_is_concrete():
    d = data(example('accounts-streaming.json'))

    assert compute(d.amount.sum()) == 100 - 200 + 300 + 400 - 500
    assert compute(d.amount.sum()) == 100 - 200 + 300 + 400 - 500


def test_repr_on_nd_array_doesnt_err():
    d = data(np.ones((2, 2, 2)))
    expr_repr(d + 1)


def test_generator_reprs_concretely():
    x = [1, 2, 3, 4, 5, 6]
    d = data(x)
    expr = d[d > 2] + 1
    assert '4' in expr_repr(expr)


def test_incompatible_types():
    d = data(pd.DataFrame(L, columns=['id', 'name', 'amount']))

    with pytest.raises(ValueError):
        d.id == 'foo'

    result = compute(d.id == 3)
    expected = pd.Series([False, False, True, False, False], name='id')
    tm.assert_series_equal(result, expected)


def test___array__():
    x = np.ones(4)
    d = data(x)
    assert (np.array(d + 1) == x + 1).all()

    d = data(x[:2])
    x[2:] = d + 1
    assert x.tolist() == [1, 1, 2, 2]


def test_python_scalar_protocols():
    d = data(1)
    assert int(d + 1) == 2
    assert float(d + 1.0) == 2.0
    assert bool(d > 0) is True
    assert complex(d + 1.0j) == 1 + 1.0j


def test_iter():
    x = np.ones(4)
    d = data(x)
    assert list(d + 1) == [2, 2, 2, 2]


@pytest.mark.xfail(
    reason="DataFrame constructor doesn't yet support __array__"
)
def test_DataFrame():
    x = np.array([(1, 2), (1., 2.)], dtype=[('a', 'i4'), ('b', 'f4')])
    d = data(x)
    assert isinstance(pd.DataFrame(d), pd.DataFrame)


def test_head_compute():
    tdata = tm.makeMixedDataFrame()
    t = symbol('t', discover(tdata))
    db = into('sqlite:///:memory:::t', tdata, dshape=t.dshape)
    n = 2
    d = data(db)

    # skip the header and the ... at the end of the repr
    expr = d.head(n)
    s = expr_repr(expr)
    assert '...' not in s
    result = s.split('\n')[1:]
    assert len(result) == n


def test_scalar_sql_compute():
    t = into('sqlite:///:memory:::t', tdata,
             dshape=dshape('var * {name: string, amount: int}'))
    d = data(t)
    assert expr_repr(d.amount.sum()) == '300'


def test_no_name_for_simple_data():
    d = data([1, 2, 3])
    assert expr_repr(d) == '    \n0  1\n1  2\n2  3'
    assert not d._name

    d = data(1)
    assert not d._name
    assert expr_repr(d) == '1'


def test_coerce_date_and_datetime():
    x = datetime.datetime.now().date()
    d = data(x)
    assert expr_repr(d) == repr(x)

    x = pd.Timestamp.now()
    d = data(x)
    assert expr_repr(d) == repr(x)

    x = np.nan
    d = data(x, dshape='datetime')
    assert expr_repr(d) == repr(pd.NaT)

    x = float('nan')
    d = data(x, dshape='datetime')
    assert expr_repr(d) == repr(pd.NaT)


def test_coerce_timedelta():
    x = datetime.timedelta(days=1, hours=2, minutes=3)
    d = data(x)

    assert expr_repr(d) == repr(x)


def test_highly_nested_repr():
    tdata = [[0, [[1, 2], [3]], 'abc']]
    d = data(tdata)
    assert 'abc' in expr_repr(d.head())


def test_asarray_fails_on_different_column_names():
    vs = {'first': [2., 5., 3.],
          'second': [4., 1., 4.],
          'third': [6., 4., 3.]}
    df = pd.DataFrame(vs)
    with pytest.raises(ValueError) as excinfo:
        data(df, fields=list('abc'))

    inmsg = ("data(data_source).relabel(first='a', second='b', third='c')"
             " to rename")
    assert inmsg in str(excinfo.value)


def test_functions_as_bound_methods():
    """
    Test that all functions on a _Data object are instance methods
    of that object.
    """
    # Filter out __class__ and friends that are special, these can be
    # callables without being instance methods.
    callable_attrs = filter(
        callable, (
            getattr(t, a, None)
            for a in dir(t)
            if not (a.startswith('__') or a.endswith('__'))
        ),
    )
    for attr in callable_attrs:
        assert isinstance(attr, MethodType)
        # Make sure this is bound to the correct object.
        assert attr.__self__ is t or attr.__self__ is type(t)  # classmethod


def test_all_string_infer_header():
    sdata = """x,tl,z
Be careful driving.,hy,en
Be careful.,hy,en
Can you translate this for me?,hy,en
Chicago is very different from Boston.,hy,en
Don't worry.,hy,en"""
    with tmpfile('.csv') as fn:
        with open(fn, 'w') as f:
            f.write(sdata)

        tdata = data(fn, has_header=True)
        assert tdata.data.has_header
        assert tdata.fields == ['x', 'tl', 'z']


def test_csv_with_trailing_commas():
    with tmpfile('.csv') as fn:
        with open(fn, 'wt') as f:
            # note the trailing space in the header
            f.write('a,b,c, \n1, 2, 3, ')
        csv = CSV(fn)
        assert expr_repr(data(fn))
        assert discover(csv).measure.names == [
            'a', 'b', 'c', ''
        ]
    with tmpfile('.csv') as fn:
        with open(fn, 'wt') as f:
            f.write('a,b,c,\n1, 2, 3, ')  # NO trailing space in the header
        csv = CSV(fn)
        assert expr_repr(data(fn))
        assert discover(csv).measure.names == [
            'a', 'b', 'c', 'Unnamed: 3'
        ]


def test_pickle_roundtrip():
    ds = data(1)
    assert ds.isidentical(
        pickle.loads(pickle.dumps(ds, protocol=pickle.HIGHEST_PROTOCOL)),
    )
    assert (ds + 1).isidentical(
        pickle.loads(pickle.dumps(ds + 1, protocol=pickle.HIGHEST_PROTOCOL)),
    )
    es = data(np.array([1, 2, 3]))
    rs = pickle.loads(pickle.dumps(es, protocol=pickle.HIGHEST_PROTOCOL))
    assert (es.data == rs.data).all()
    assert_dshape_equal(es.dshape, rs.dshape)


def test_nameless_data():
    tdata = [('a', 1)]
    assert repr(tdata) in expr_repr(data(tdata))


def test_partially_bound_expr():
    df = pd.DataFrame([(1, 'Alice', 100),
                       (2, 'Bob', -200),
                       (3, 'Charlie', 300),
                       (4, 'Denis', 400),
                       (5, 'Edith', -500)],
                      columns=['id', 'name', 'balance'])
    tdata = data(df, name='data')
    a = symbol('a', 'int')
    expr = tdata.name[tdata.balance > a]
    assert expr_repr(expr) == 'data[data.balance > a].name'


def test_isidentical_regr():
    # regression test for #1387
    tdata = np.array([(np.nan,), (np.nan,)], dtype=[('a', 'float64')])
    ds = data(tdata)
    assert ds.a.isidentical(ds.a)


@pytest.mark.parametrize('data,dshape,exp_type',
                         [(1, symbol('x', 'int').dshape, int),
                          # test 1-d to series
                          (into(da.core.Array, [1, 2], chunks=(10,)),
                           dshape('2 * int'),
                           pd.Series),
                          # test 2-d tabular to dataframe
                          (into(da.core.Array,
                                [{'a': 1, 'b': 2}, {'a': 3, 'b': 4}],
                                chunks=(10,)),
                           dshape('2 * {a: int, b: int}'),
                           pd.DataFrame),
                          # test 2-d non tabular to ndarray
                          (into(da.core.Array,
                                [[1, 2], [3, 4]],
                                chunks=(10, 10)),
                           dshape('2 *  2 * int'),
                           np.ndarray)])
def test_coerce_core(data, dshape, exp_type):
    assert isinstance(coerce_core(data, dshape), exp_type)


@pytest.mark.parametrize('data,res',
                         [(1, True),
                          (1.1, True),
                          ("foo", True),
                          ([1, 2], False),
                          ((1, 2), False),
                          (pd.Series([1, 2]), False)])
def test_iscorescalar(data, res):
    assert iscorescalar(data) == res


@pytest.mark.parametrize('data,res',
                         [(1, False),
                          ("foo", False),
                          ([1, 2], True),
                          ((1, 2), True),
                          (pd.Series([1, 2]), True),
                          (pd.DataFrame([[1, 2], [3, 4]]), True),
                          (np.ndarray([1, 2]), True),
                          (into(da.core.Array, [1, 2], chunks=(10,)), False)])
def test_iscoresequence(data, res):
    assert iscoresequence(data) == res


@pytest.mark.parametrize('data,res',
                         [(1, True),
                          ("foo", True),
                          ([1, 2], True),
                          ((1, 2), True),
                          (pd.Series([1, 2]), True),
                          (pd.DataFrame([[1, 2], [3, 4]]), True),
                          (np.ndarray([1, 2]), True),
                          (into(da.core.Array, [1, 2], chunks=(10,)), False)])
def test_iscoretype(data, res):
    assert iscoretype(data) == res



from __future__ import absolute_import, division, print_function

from .compatibility import basestring
from .compute.mongo import dispatch

try:
    from pymongo.collection import Collection
    from pymongo import ASCENDING
except ImportError:
    Collection = type(None)


__all__ = ['create_index']


@dispatch(object)
def scrub_keys(o):
    """Add an ascending sort key when pass a string, to make the MongoDB
    interface similar to SQL.
    """
    raise NotImplementedError("scrub_keys not implemented for type %r" %
                              type(o).__name__)


@dispatch(basestring)
def scrub_keys(s):
    return s, ASCENDING


@dispatch(tuple)
def scrub_keys(t):
    return t


@dispatch(list)
def scrub_keys(seq):
    for el in seq:
        if not isinstance(el, (tuple, basestring)):
            raise TypeError('indexing keys must be a string or pair of '
                            '(<column name>, <parameter>)')
        yield scrub_keys(el)


@dispatch(Collection, basestring)
def create_index(coll, key, **kwargs):
    coll.create_index(key, **kwargs)


@dispatch(Collection, list)
def create_index(coll, keys, **kwargs):
    coll.create_index(list(scrub_keys(keys)), **kwargs)



from itertools import product

import pytest

from blaze.expr import symbol, summary
from datashape import dshape
from datashape.util.testing import assert_dshape_equal


def test_reduction_dshape():
    x = symbol('x', '5 * 3 * float32')
    assert x.sum().dshape == dshape('float64')
    assert x.sum(axis=0).dshape == dshape('3 * float64')
    assert x.sum(axis=1).dshape == dshape('5 * float64')
    assert x.sum(axis=(0, 1)).dshape == dshape('float64')


def test_keepdims():
    x = symbol('x', '5 * 3 * float32')
    assert x.sum(axis=0, keepdims=True).dshape == dshape('1 * 3 * float64')
    assert x.sum(axis=1, keepdims=True).dshape == dshape('5 * 1 * float64')
    assert x.sum(axis=(0, 1), keepdims=True).dshape == dshape(
        '1 * 1 * float64')

    assert x.std(axis=0, keepdims=True).shape == (1, 3)


def test_summary_keepdims():
    x = symbol('x', '5 * 3 * float32')
    assert summary(a=x.min(), b=x.max()).dshape == \
        dshape('{a: float32, b: float32}')
    assert summary(a=x.min(), b=x.max(), keepdims=True).dshape == \
        dshape('1 * 1 * {a: float32, b: float32}')


def test_summary_axis():
    x = symbol('x', '5 * 3 * float32')
    assert summary(a=x.min(), b=x.max(), axis=0).dshape == \
        dshape('3 * {a: float32, b: float32}')
    assert summary(a=x.min(), b=x.max(), axis=1).dshape == \
        dshape('5 * {a: float32, b: float32}')
    assert summary(a=x.min(), b=x.max(), axis=1, keepdims=True).dshape == \
        dshape('5 * 1 * {a: float32, b: float32}')


def test_summary_str():
    x = symbol('x', '5 * 3 * float32')
    assert 'keepdims' not in str(summary(a=x.min(), b=x.max()))


def test_axis_kwarg_is_normalized_to_tuple():
    x = symbol('x', '5 * 3 * float32')
    exprs = [x.sum(), x.sum(axis=1), x.sum(axis=[1]), x.std(), x.mean(axis=1)]
    for expr in exprs:
        assert isinstance(expr.axis, tuple)


def test_summary_with_multiple_children():
    t = symbol('t', 'var * {x: int, y: int, z: int}')

    assert summary(a=t.x.sum() + t.y.sum())._child.isidentical(t)


def test_dir():
    t = symbol('t', '10 * int')
    assert 'mean' in dir(t)

    t = symbol('t', 'int')
    assert 'mean' not in dir(t)


def test_norms():
    x = symbol('x', '5 * 3 * float32')
    assert x.vnorm().isidentical(x.vnorm('fro'))
    assert x.vnorm().isidentical(x.vnorm(2))
    assert x.vnorm(axis=0).shape == (3,)
    assert x.vnorm(axis=0, keepdims=True).shape == (1, 3)


@pytest.mark.parametrize('reduc', ['max', 'min', 'sum', 'mean', 'std', 'var'])
def test_reductions_on_record_dshape(reduc):
    t = symbol('t', '10 * {a: int64, b: string}')
    with pytest.raises(AttributeError):
        getattr(t, reduc)


@pytest.mark.parametrize('reduc', ['max', 'min', 'sum', 'mean', 'std', 'var'])
def test_boolean_has_reductions(reduc):
    assert hasattr(symbol('t', 'var * bool'), reduc)


@pytest.mark.parametrize(['reduc', 'measure'],
                         product(['max', 'min'],
                                 ['date', 'datetime', 'timedelta']))
def test_max_min_on_datetime_and_timedelta(reduc, measure):
    assert hasattr(symbol('t', 'var * %s' % measure), reduc)


def test_reduction_naming_with_generated_leaves():
    assert symbol('_', 'var * float64').sum()._name == 'sum'


@pytest.mark.parametrize('func', ['sum', 'mean', 'std', 'var'])
def test_decimal_reduction(func):
    t = symbol('t', 'var * decimal[11, 2]')
    method = getattr(t, func)
    assert_dshape_equal(method().dshape, dshape("decimal[11, 2]"))


@pytest.mark.parametrize('func', ['sum', 'mean', 'std', 'var'])
def test_timedelta_reduction(func):
    t = symbol('t', 'var * timedelta')
    method = getattr(t, func)
    assert_dshape_equal(method().dshape, dshape("timedelta"))



""" Abstraction around partitionable data

Some data can be partitioned into disjoint pieces.
Examples include lists, tuples, and arrays of various flavors.

This module provides functions to abstract away the handling of such
partitions.  It proposes three operations, each of these operations deals with
data (like a numpy array) and a *partition-key* which is some token to define a
part of the data.  In the case of numpy (and other array-like data) a natural
partition-key might be a tuple of slices, e.g.

    (slice(0, 10), slice(0, 10)) :: partition-key

Our three operations:


    partitions :: data -> [partition-keys]
        Get list or list-of-lists of partition-keys

    partition_get :: data, partition-key -> partition
        Get a particular partition from a dataset

    partition_set :: data, partition-key, value -> void
        Set the value of a particular partition in a dataset

Using these three operations we should be able to write down very simple and
abstract algorithms like "copy"

    def copy(in_data, out_data):
        for in_key, out_key in zip(partitions(in_data), partitions(out_data)):
            data = partition_get(in_data, in_key)
            partition_set(out_data, out_key, data)
"""
from __future__ import absolute_import, division, print_function

import numpy as np
from math import ceil
import toolz
import itertools
try:
    import h5py
    from h5py import Dataset
except ImportError:
    Dataset = type(None)

try:
    from bcolz import carray, ctable
except ImportError:
    carray = ctable = type(None)

from .dispatch import dispatch

Array = (np.ndarray, Dataset, ctable, carray)

@dispatch(Array, object)
def partition_get(data, part, chunksize=None):
    return data[part]


@dispatch(Array, object, object)
def partition_set(data, part, value, chunksize=None, keepdims=True):
    if not keepdims and 1 in value.shape:
        value = value.squeeze()
    data[part] = value
    return data


def slices1d(n, k, keepdims=False):
    """

    >>> slices1d(10, 5)
    [slice(0, 5, None), slice(5, 10, None)]

    >>> slices1d(10, 6)
    [slice(0, 6, None), slice(6, 10, None)]

    Ordinarily single dimensions are flattened down
    >>> slices1d(10, 1)
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    Reverse this behavior with the keepdims keyword argument
    >>> slices1d(3, 1, keepdims=True)
    [slice(0, 1, None), slice(1, 2, None), slice(2, 3, None)]
    """
    if k == 1 and not keepdims:
        return list(range(n))
    return [slice(k*i, min(k*(i + 1), n)) for i in range(int(ceil(float(n)/k)))]



def tuplepack(x):
    if isinstance(x, (tuple, list)):
        return tuple(x)
    else:
        return (x,)


@dispatch(Array)
def partitions(data, chunksize=None, keepdims=False):
    per_dim = [slices1d(s, c, keepdims=keepdims)
                    for s, c in zip(data.shape, chunksize)]
    return itertools.product(*per_dim)


def flatten(x):
    """

    >>> flatten([[1]])
    [1]
    >>> flatten([[1, 2], [3, 4]])
    [1, 2, 3, 4]
    >>> flatten([[[1], [2]], [[3], [4]]])
    [1, 2, 3, 4]
    """
    if isinstance(x[0], list):
        return list(toolz.concat(map(flatten, x)))
    else:
        return x



from __future__ import absolute_import, division, print_function

from datetime import datetime, date, timedelta
import sys


def identity(x):
    return x


def asday(dt):
    if isinstance(dt, datetime):
        return dt.date()
    else:
        return dt


def asweek(dt):
    if isinstance(dt, datetime):
        dt = dt.date()
    return dt - timedelta(days=dt.isoweekday() - 1)


def ashour(dt):
    return datetime(dt.year, dt.month, dt.day, dt.hour, tzinfo=dt.tzinfo)


def asminute(dt):
    return datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute,
                    tzinfo=dt.tzinfo)


def assecond(dt):
    return datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute,
                    dt.second, tzinfo=dt.tzinfo)


def asmillisecond(dt):
    return datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute,
                    dt.second, dt.microsecond // 1000, tzinfo=dt.tzinfo)

if sys.version_info < (2, 7):
    def total_seconds(td):
        """ Total seconds of a timedelta

        For Python 2.6 compatibility
        """
        return (td.microseconds + 1e6 * (td.seconds + 24 * 3600 * td.days)) / 1e6
else:
    total_seconds = timedelta.total_seconds


unit_map = {'year': 'asyear',
            'month': 'asmonth',
            'week': 'asweek',
            'day': 'asday',
            'hour': 'ashour',
            'minute': 'asminute',
            'second': 'assecond',
            'millisecond': 'asmillisecond',
            'microsecond': identity}


def truncate_year(dt, measure):
    """
    Truncate by years

    >>> dt = datetime(2003, 6, 25, 12, 30, 0)
    >>> truncate_year(dt, 1)
    datetime.date(2003, 1, 1)

    >>> truncate_year(dt, 5)
    datetime.date(2000, 1, 1)
    """
    return date(dt.year // measure * measure, 1, 1)


def truncate_month(dt, measure):
    """
    Truncate by months

    >>> dt = datetime(2000, 10, 25, 12, 30, 0)
    >>> truncate_month(dt, 1)
    datetime.date(2000, 10, 1)

    >>> truncate_month(dt, 4)
    datetime.date(2000, 8, 1)
    """
    months = dt.year * 12 + dt.month
    months = months // measure * measure
    return date((months - 1) // 12, (months - 1) % 12 + 1, 1)


def truncate_day(dt, measure):
    """
    Truncate by days

    >>> dt = datetime(2000, 6, 27, 12, 30, 0)
    >>> truncate_day(dt, 1)
    datetime.date(2000, 6, 27)
    >>> truncate_day(dt, 3)
    datetime.date(2000, 6, 25)

    """
    days = dt.toordinal()
    days = days // measure * measure
    return date.fromordinal(days)

oneday = timedelta(days=1)


def truncate_week(dt, measure):
    """
    Truncate by weeks

    >>> dt = datetime(2000, 6, 22, 12, 30, 0)
    >>> truncate_week(dt, 1)
    datetime.date(2000, 6, 18)
    >>> truncate_week(dt, 3)
    datetime.date(2000, 6, 4)

    Weeks are defined by having isoweekday == 7 (Sunday)
    >>> truncate_week(dt, 1).isoweekday()
    7
    """
    return truncate_day(dt, measure * 7)


epoch = datetime.utcfromtimestamp(0)


def utctotimestamp(dt):
    """
    Convert a timestamp to seconds

    >>> dt = datetime(2000, 1, 1)
    >>> utctotimestamp(dt)
    946684800.0

    >>> datetime.utcfromtimestamp(946684800)
    datetime.datetime(2000, 1, 1, 0, 0)
    """
    return total_seconds(dt - epoch)


def truncate_minute(dt, measure):
    """
    Truncate by minute

    >>> dt = datetime(2000, 1, 1, 12, 30, 38)
    >>> truncate_minute(dt, 1)
    datetime.datetime(2000, 1, 1, 12, 30)
    >>> truncate_minute(dt, 12)
    datetime.datetime(2000, 1, 1, 12, 24)
    """
    return asminute(truncate_second(dt, measure * 60))


def truncate_hour(dt, measure):
    """
    Truncate by hour

    >>> dt = datetime(2000, 1, 1, 12, 30, 38)
    >>> truncate_hour(dt, 1)
    datetime.datetime(2000, 1, 1, 12, 0)
    >>> truncate_hour(dt, 5)
    datetime.datetime(2000, 1, 1, 10, 0)
    """
    return ashour(truncate_second(dt, measure * 3600))


def truncate_second(dt, measure):
    """
    Truncate by second

    >>> dt = datetime(2000, 1, 1, 12, 30, 38)
    >>> truncate_second(dt, 15)
    datetime.datetime(2000, 1, 1, 12, 30, 30)
    """
    d = datetime(
        dt.year, dt.month, dt.day, tzinfo=dt.tzinfo)  # local zero for seconds
    seconds = total_seconds(dt - d) // measure * measure
    return dt.utcfromtimestamp(seconds + utctotimestamp(d))


def truncate_millisecond(dt, measure):
    """
    Truncate by millisecond

    >>> dt = datetime(2000, 1, 1, 12, 30, 38, 12345)
    >>> truncate_millisecond(dt, 5)
    datetime.datetime(2000, 1, 1, 12, 30, 38, 10000)
    """
    d = datetime(
        dt.year, dt.month, dt.day, tzinfo=dt.tzinfo)  # local zero for seconds
    seconds = total_seconds(dt - d) * 1000 // measure * measure / 1000. + 1e-7
    return dt.utcfromtimestamp(seconds + utctotimestamp(d))


def truncate_microsecond(dt, measure):
    """
    Truncate by microsecond

    >>> dt = datetime(2000, 1, 1, 12, 30, 38, 12345)
    >>> truncate_microsecond(dt, 100)
    datetime.datetime(2000, 1, 1, 12, 30, 38, 12300)
    """
    d = datetime(
        dt.year, dt.month, dt.day, tzinfo=dt.tzinfo)  # local zero for seconds
    seconds = total_seconds(dt - d) * 1000000 // measure * measure / 1000000
    return dt.utcfromtimestamp(seconds + utctotimestamp(d))


truncate_functions = {'year': truncate_year,
                      'month': truncate_month,
                      'week': truncate_week,
                      'day': truncate_day,
                      'hour': truncate_hour,
                      'minute': truncate_minute,
                      'second': truncate_second,
                      'millisecond': truncate_millisecond,
                      'microsecond': truncate_microsecond}


def truncate(dt, measure, unit):
    """ Truncate datetimes

    Examples
    --------

    >>> dt = datetime(2003, 6, 25, 12, 30, 0)
    >>> truncate(dt, 1, 'day')
    datetime.date(2003, 6, 25)

    >>> truncate(dt, 5, 'hours')
    datetime.datetime(2003, 6, 25, 10, 0)

    >>> truncate(dt, 3, 'months')
    datetime.date(2003, 6, 1)
    """
    from blaze.expr.datetime import normalize_time_unit
    unit = normalize_time_unit(unit)
    return truncate_functions[unit](dt, measure)



from __future__ import absolute_import, division, print_function

from blaze.utils import example
from odo.backends.json import JSON, JSONLines
from odo import Chunks
from blaze import symbol, discover, compute, into, data


js = JSON(example('accounts.json'))
jss = JSONLines(example('accounts-streaming.json'))

s = symbol('s', discover(js))


def test_simple():
    assert compute(s.count(), js) == 5
    assert compute(s.count(), jss) == 5


def test_less_simple():
    assert compute(s.amount.sum(), js) == 100
    assert compute(s.amount.sum(), jss) == 100


def test_chunks_json():
    r = data(example('accounts-streaming*.json'))
    assert isinstance(r.data, Chunks)
    assert compute(s.amount.sum(), r) == 200



from __future__ import absolute_import, division, print_function

from .object_hook import object_hook, register
import werkzeug.exceptions as wz_ex
from blaze.compatibility import builtins, reduce
import numpy as np
import pandas as pd

_converters_trusted = object_hook._converters.copy()


def object_hook_trusted(ob, _converters=_converters_trusted):
    return object_hook(ob, _converters=_converters)
object_hook_trusted._converters = _converters_trusted
object_hook_trusted.register = register(converters=_converters_trusted)
del _converters_trusted


@object_hook_trusted.register('callable')
def numpy_pandas_function_from_str(f):
    """
    reconstruct function from string representation
    """
    if f.startswith(np.__name__):
        mod = np
    elif f.startswith(pd.__name__):
        mod = pd
    elif f.startswith(builtins.__name__):
        mod = builtins
    else:
        msg = ("Function {} not recognized; only numpy, pandas, or builtin "
               "functions are supported.")
        raise wz_ex.NotImplemented(msg.format(f))
    fcn = reduce(getattr, f.split('.')[1:], mod)
    return fcn



from __future__ import absolute_import, division, print_function

import types

import datashape
from datashape import dshape, var, datetime_, float32, int64, bool_
from datashape.util.testing import assert_dshape_equal
import pandas as pd
import pytest

from blaze.compatibility import pickle
from blaze.expr import (
    Expr,
    Field,
    Node,
    coalesce,
    label,
    symbol,
    transform,
)


def test_arguments():
    assert Expr._arguments == ('_child',)
    assert Node._arguments == ('_child',)


def test_Symbol():
    e = symbol('e', '3 * 5 * {name: string, amount: int}')
    assert e.dshape == dshape('3 * 5 * {name: string, amount: int}')
    assert e.shape == (3, 5)
    assert str(e) == 'e'


def test_symbol_caches():
    assert symbol('e', 'int') is symbol('e', 'int')


def test_Symbol_tokens():
    assert symbol('x', 'int').isidentical(symbol('x', 'int'))
    assert not symbol('x', 'int').isidentical(symbol('x', 'int', 1))


def test_Field():
    e = symbol('e', '3 * 5 * {name: string, amount: int}')
    assert 'name' in dir(e)
    assert e.name.dshape == dshape('3 * 5 * string')
    assert e.name.schema == dshape('string')
    assert e.amount._name == 'amount'


def test_nested_fields():
    e = symbol(
        'e', '3 * {name: string, payments: var * {amount: int, when: datetime}}')
    assert e.payments.dshape == dshape(
        '3 * var * {amount: int, when: datetime}')
    assert e.payments.schema == dshape('{amount: int, when: datetime}')
    assert 'amount' in dir(e.payments)
    assert e.payments.amount.dshape == dshape('3 * var * int')


@pytest.mark.xfail
def test_partialed_methods_have_docstrings():
    e = symbol('e', '3 * 5 * {name: string, amount: int}')
    assert 'string comparison' in e.name.str.like.__doc__


def test_relabel():
    e = symbol('e', '{name: string, amount: int}')
    assert e.relabel(amount='balance').fields == ['name', 'balance']


def test_meaningless_relabel_doesnt_change_input():
    e = symbol('e', '{name: string, amount: int}')
    assert e.relabel(amount='amount').isidentical(e)


def test_relabel_with_invalid_identifiers_reprs_as_dict():
    s = symbol('s', '{"0": int64}')
    assert str(s.relabel({'0': 'foo'})) == "s.relabel({'0': 'foo'})"


def test_dir():
    e = symbol('e', '3 * 5 * {name: string, amount: int, x: real}')

    assert 'name' in dir(e)
    assert 'name' not in dir(e.name)
    assert 'isnan' in dir(e.x)
    assert 'isnan' not in dir(e.amount)


def test_label():
    e = symbol('e', '3 * int')
    assert e._name == 'e'
    assert label(e, 'foo')._name == 'foo'
    assert label(e, 'e').isidentical(e)


def test_fields_with_spaces():
    e = symbol('e', '{x: int, "a b": int}')
    assert isinstance(e['a b'], Field)
    assert 'a b' not in dir(e)

    assert 'a_b' in dir(e)
    assert e.a_b.isidentical(e['a b'])


def test_selection_name_matches_child():
    t = symbol('t', 'var * {x: int, "a.b": int}')
    assert t.x[t.x > 0]._name == t.x._name
    assert t.x[t.x > 0].fields == t.x.fields


def test_symbol_subs():
    assert symbol('e', '{x: int}').isidentical(symbol('e', '{x: int}', None))
    assert symbol('e', '{x: int}').isidentical(symbol('e', dshape('{x: int}')))
    e = symbol('e', '{x: int, y: int}')
    f = symbol('f', '{x: int, y: int}')
    d = {'e': 'f'}
    assert e._subs(d).isidentical(f)


def test_multiple_renames_on_series_fails():
    t = symbol('s', 'var * {timestamp: datetime}')
    with pytest.raises(ValueError):
        t.timestamp.relabel(timestamp='date', hello='world')


def test_map_with_rename():
    t = symbol('s', 'var * {timestamp: datetime}')
    result = t.timestamp.map(lambda x: x.date(), schema='{date: datetime}')
    with pytest.raises(ValueError):
        result.relabel(timestamp='date')
    assert result.fields == ['date']


def test_non_option_does_not_have_notnull():
    s = symbol('s', '5 * int32')
    assert not hasattr(s, 'notnull')


def test_notnull_dshape():
    assert symbol('s', '5 * ?int32').notnull().dshape == 5 * bool_


def test_hash_to_different_values():
    s = symbol('s', var * datetime_)
    expr = s >= pd.Timestamp('20121001')
    expr2 = s >= '20121001'
    assert expr2 & expr is not None
    assert hash(expr) == hash(expr2)


def test_hash():
    e = symbol('e', 'int')
    h = hash(e)
    assert isinstance(h, int)
    assert h == hash(e)

    assert hash(symbol('e', 'int')) == hash(symbol('e', 'int'))

    f = symbol('f', 'int')
    assert hash(e) != hash(f)

    assert hash(e._subs({'e': 'f'})) != hash(e)
    assert hash(e._subs({'e': 'f'})) == hash(f)


@pytest.mark.parametrize('dshape', [var * float32,
                                    dshape('var * float32'),
                                    'var * float32'])
def test_coerce(dshape):
    s = symbol('s', dshape)
    expr = s.coerce('int64')
    assert str(expr) == "s.coerce(to='int64')"
    assert expr.dshape == var * int64
    assert expr.schema == datashape.dshape('int64')
    assert expr.schema == expr.to


@pytest.mark.xfail(raises=AttributeError, reason='Should this be valid?')
def test_coerce_record():
    s = symbol('s', 'var * {a: int64, b: float64}')
    expr = s.coerce('{a: float64, b: float32}')
    assert str(expr) == "s.coerce(to='{a: float64, b: float32}')"


def test_method_before_name():
    t = symbol('t', 'var * {isin: int64, max: float64, count: int64}')
    assert isinstance(t['isin'], Field)
    assert isinstance(t['max'], Field)
    assert isinstance(t.max, Field)
    assert isinstance(t.isin, Field)
    assert isinstance(t['isin'].isin, types.MethodType)
    assert isinstance(t['max'].max, types.MethodType)
    assert isinstance(t.max.max, types.MethodType)
    assert isinstance(t.isin.isin, types.MethodType)
    with pytest.raises(AttributeError):
        t.count.max()


def test_pickle_roundtrip():
    t = symbol('t', 'var * int64')
    expr = (t + 1).mean()  # some expression with more than one node.
    assert expr.isidentical(pickle.loads(
        pickle.dumps(expr, protocol=pickle.HIGHEST_PROTOCOL),
    ))


def test_coalesce():
    # check case where lhs is not optional
    s = symbol('s', 'int32')
    t = symbol('t', 'int32')
    expr = coalesce(s, t)
    assert expr.isidentical(s)

    s_expr = s + s
    t_expr = t * 3
    expr = coalesce(s_expr, t_expr)
    assert expr.isidentical(s_expr)

    a = symbol('a', 'string')
    b = symbol('b', 'string')
    expr = coalesce(a, b)
    assert expr.isidentical(a)

    a_expr = a + a
    b_expr = b * 3
    expr = coalesce(a_expr, b_expr)
    assert expr.isidentical(a_expr)

    c = symbol('c', '{a: int32, b: int32}')
    d = symbol('d', '{a: int32, b: int32}')
    expr = coalesce(c, d)
    assert expr.isidentical(c)

    c_expr = transform(c, a=c.a + 1)
    d_expr = transform(d, a=d.a * 3)
    expr = coalesce(c_expr, d_expr)
    assert expr.isidentical(c_expr)

    # check case where lhs is null dshape
    u = symbol('u', 'null')
    expr = coalesce(u, s)
    assert expr.isidentical(s)

    expr = coalesce(u, a)
    assert expr.isidentical(a)

    expr = coalesce(u, c)
    assert expr.isidentical(c)

    # check optional lhs non-optional rhs
    v = symbol('v', '?int32')
    expr = coalesce(v, s)
    # rhs is not optional so the expression cannot be null
    assert_dshape_equal(expr.dshape, dshape('int32'))
    assert expr.lhs.isidentical(v)
    assert expr.rhs.isidentical(s)

    e = symbol('e', '?string')
    expr = coalesce(e, a)
    assert_dshape_equal(expr.dshape, dshape('string'))
    assert expr.lhs.isidentical(e)
    assert expr.rhs.isidentical(a)

    f = symbol('f', '?{a: int32, b: int32}')
    expr = coalesce(f, c)
    assert_dshape_equal(expr.dshape, dshape('{a: int32, b: int32}'))
    assert expr.lhs.isidentical(f)
    assert expr.rhs.isidentical(c)

    # check optional lhs non-optional rhs with promotion
    w = symbol('w', 'int64')
    expr = coalesce(v, w)
    # rhs is not optional so the expression cannot be null
    # there are no either types in datashape so we are a type large enough
    # to hold either result
    assert_dshape_equal(expr.dshape, dshape('int64'))
    assert expr.lhs.isidentical(v)
    assert expr.rhs.isidentical(w)

    # check optional lhs and rhs
    x = symbol('x', '?int32')
    expr = coalesce(v, x)
    # rhs and lhs are optional so this might be null
    assert_dshape_equal(expr.dshape, dshape('?int32'))
    assert expr.lhs.isidentical(v)
    assert expr.rhs.isidentical(x)

    # check optional lhs and rhs with promotion
    y = symbol('y', '?int64')
    expr = coalesce(v, y)
    # rhs and lhs are optional so this might be null
    # there are no either types in datashape so we are a type large enough
    # to hold either result
    assert_dshape_equal(expr.dshape, dshape('?int64'))
    assert expr.lhs.isidentical(v)
    assert expr.rhs.isidentical(y)


@pytest.mark.xfail(TypeError, reason='currently invalid type promotion')
@pytest.mark.parametrize('lhs,rhs,expected', (
    ('?{a: int32}', '{a: int64}', '{a: int64}'),
    ('?{a: int32}', '?{a: int64}', '?{a: int64}'),
))
def test_coalesce_invalid_promotion(lhs, rhs, expected):
    # Joe 2016-03-16: imho promote(record, record) should check that the keys
    # are the same and then create a new record from:
    # zip(keys, map(promote, lhs, rhs))
    f = symbol('e', lhs)
    g = symbol('g', rhs)
    expr = coalesce(f, g)
    assert_dshape_equal(expr.dshape, dshape(expected))
    assert expr.lhs.isidentical(f)
    assert expr.rhs.isidentical(g)


def test_cast():
    s = symbol('s', 'int32')

    assert_dshape_equal(s.cast('int64').dshape, dshape('int64'))
    assert_dshape_equal(s.cast(dshape('int64')).dshape, dshape('int64'))
    assert_dshape_equal(s.cast('var * int32').dshape, dshape('var * int32'))
    assert_dshape_equal(
        s.cast(dshape('var * int64')).dshape,
        dshape('var * int64'),
    )
    assert_dshape_equal(s.cast('var * int64').dshape, dshape('var * int64'))
    assert_dshape_equal(
        s.cast(dshape('var * int64')).dshape,
        dshape('var * int64'),
    )



import pytest

nb = pytest.importorskip('numba')

from numba import float64, int64, float32
from numba.types import NPDatetime as datetime64, NPTimedelta as timedelta64
from numba.types import CharSeq as char, UnicodeCharSeq as unichar
from blaze.compute.numba import compute_signature, get_numba_type
from blaze import symbol
import datashape


def test_compute_signature():
    s = symbol('s', 'int64')
    t = symbol('t', 'float32')
    d = symbol('d', 'datetime')

    assert compute_signature(s + t) == float64(int64, float32)
    assert (compute_signature(d.truncate(days=1)) ==
            datetime64('D')(datetime64('us')))
    assert compute_signature(d.day + 1) == int64(datetime64('us'))


def test_get_numba_type():
    assert get_numba_type(datashape.bool_) == nb.bool_
    assert get_numba_type(datashape.date_) == datetime64('D')
    assert get_numba_type(datashape.datetime_) == datetime64('us')
    assert get_numba_type(datashape.timedelta_) == timedelta64('us')
    assert get_numba_type(datashape.TimeDelta('D')) == timedelta64('D')
    assert get_numba_type(datashape.int64) == int64
    assert get_numba_type(datashape.String(7, "A")) == char(7)
    assert get_numba_type(datashape.String(None, "A")) == nb.types.string
    assert get_numba_type(datashape.String(7)) == unichar(7)


def test_fail_on_object_type():
    with pytest.raises(TypeError):
        get_numba_type(datashape.object_)


@pytest.mark.xfail(raises=TypeError,
                   reason="Cannot infer variable length string type yet")
def test_get_numba_type_failures():
    get_numba_type(datashape.string)


@pytest.mark.xfail(raises=TypeError,
                   reason='Cannot infer type of record dshapes yet')
def test_get_record_type():
    get_numba_type(datashape.dshape('10 * {a: int64}'))



from __future__ import absolute_import, division, print_function

import pytest

import numpy as np
from pandas import DataFrame
from odo import into
from datashape.predicates import isscalar, iscollection, isrecord
from blaze.expr import symbol, by
from blaze.interactive import data
from blaze.compute import compute
from blaze.expr.functions import sin, exp


t = symbol('t', 'var * {amount: int64, id: int64, name: string}')


@pytest.fixture
def sources():
    L = [[ 100, 1, 'Alice'],
         [ 200, 2, 'Bob'],
         [ 300, 3, 'Charlie'],
         [-400, 4, 'Dan'],
         [ 500, 5, 'Edith']]

    df = DataFrame(L, columns=['amount', 'id', 'name'])

    x = into(np.ndarray, df)

    try:
        import sqlalchemy
        sql = data('sqlite:///:memory:::accounts', dshape=t.dshape)
        into(sql, L)
    except:
        sql = None

    try:
        import bcolz
        bc = into(bcolz.ctable, df)
    except ImportError:
        bc = None

    try:
        import pymongo
    except ImportError:
        pymongo = mongo = None
    if pymongo:

        try:
            db = pymongo.MongoClient().db

            try:
                coll = db._test_comprehensive
            except AttributeError:
                coll = db['_test_comprehensive']

            coll.drop()
            mongo = into(coll, df)
        except pymongo.errors.ConnectionFailure:
            mongo = None

    return df, x, sql, bc, mongo


# {expr: [list-of-exclusions]}
@pytest.fixture
def expressions(sources):
    df, x, sql, bc, mongo = sources

    return {
        t: [],
        t['id']: [],
        abs(t['amount']): [],
        t.id.max(): [],
        t.amount.sum(): [],
        t.amount.sum(keepdims=True): [],
        t.amount.count(keepdims=True): [],
        t.amount.nunique(keepdims=True): [mongo],
        t.amount.nunique(): [],
        t.amount.head(): [],
        t.amount + 1: [mongo],
        sin(t.amount): [sql, mongo], # sqlite doesn't support trig
        exp(t.amount): [sql, mongo],
        t.amount > 50: [mongo],
        t[t.amount > 50]: [],
        t[t.name.str.like('Alic*')]: [],
        t.sort('name'): [bc],
        t.sort('name', ascending=False): [bc],
        t.head(3): [],
        t.name.distinct(): [],
        t[t.amount > 50]['name']: [], # odd ordering issue
        t.id.map(lambda x: x + 1, schema='int64', name='id'): [sql, mongo],
        t[t.amount > 50]['name']: [],
        by(t.name, total=t.amount.sum()): [],
        by(t.id, count=t.id.count()): [],
        by(t[['id', 'amount']], count=t.id.count()): [],
        by(t[['id', 'amount']], total=(t.amount + 1).sum()): [mongo],
        by(t[['id', 'amount']], n=t.name.nunique()): [mongo, bc],
        by(t.id, count=t.amount.count()): [],
        by(t.id, n=t.id.nunique()): [mongo, bc],
        # by(t, count=t.count()): [],
        # by(t.id, count=t.count()): [],
        t[['amount', 'id']]: [x], # https://github.com/numpy/numpy/issues/3256
        t[['id', 'amount']]: [x, bc], # bcolz sorting
        t[0]: [sql, mongo, bc],
        t[::2]: [sql, mongo, bc],
        t.id.utcfromtimestamp: [sql],
        t.distinct().nrows: [],
        t.nelements(axis=0): [],
        t.nelements(axis=None): [],
        t.amount.truncate(200): [sql]
    }


def df_eq(a, b):
    return (list(a.columns) == list(b.columns)
            # and list(a.dtypes) == list(b.dtypes)
            and into(set, into(list, a)) == into(set, into(list, b)))


def typename(obj):
    return type(obj).__name__


def test_base(expressions, sources):
    base, x, sql, bc, mongo = sources
    for expr, exclusions in expressions.items():
        if iscollection(expr.dshape):
            model = into(DataFrame, into(np.ndarray, expr._subs({t: data(base, t.dshape)})))
        else:
            model = compute(expr._subs({t: data(base, t.dshape)}))
        print('\nexpr: %s\n' % expr)
        for source in sources:
            if source is None or id(source) in map(id, exclusions):
                continue
            print('%s <- %s' % (typename(model), typename(source)))
            T = data(source)
            if iscollection(expr.dshape):
                result = into(type(model), expr._subs({t: T}))
                if isscalar(expr.dshape.measure):
                    assert set(into(list, result)) == set(into(list, model))
                else:
                    assert df_eq(result, model)
            elif isrecord(expr.dshape):
                result = compute(expr._subs({t: T}))
                assert into(tuple, result) == into(tuple, model)
            else:
                result = compute(expr._subs({t: T}))
                try:
                    result = result.scalar()
                except AttributeError:
                    pass
                assert result == model



from __future__ import absolute_import, print_function, division

import sys
import pytest

pytestmark = pytest.mark.skipif(
    sys.version_info.major == 3,
    reason="PyHive doesn't work with Python 3.x"
)

pyspark = pytest.importorskip('pyspark')
py4j = pytest.importorskip('py4j')
sa = pytest.importorskip('sqlalchemy')
pytest.importorskip('pyhive.sqlalchemy_hive')

import os
import itertools
import shutil

from functools import partial

from py4j.protocol import Py4JJavaError
import numpy as np
import pandas as pd
import pandas.util.testing as tm
from blaze import compute, symbol, into, by, sin, exp, cos, tan, join
from blaze.interactive import iscorescalar

from pyspark.sql import DataFrame as SparkDataFrame

try:
    from pyspark.sql.utils import AnalysisException
except ImportError:
    AnalysisException = Py4JJavaError

from pyspark import HiveContext
from pyspark.sql import Row
from odo import odo, discover
from odo.utils import tmpfile


data = [['Alice', 100.0, 1],
        ['Bob', 200.0, 2],
        ['Alice', 50.0, 3]]

date_data = []

np.random.seed(0)

for attr in ('YearBegin', 'MonthBegin', 'Day', 'Hour', 'Minute', 'Second'):
    rng = pd.date_range(start='now', periods=len(data),
                        freq=getattr(pd.datetools, attr)()).values
    date_data += list(zip(np.random.choice(['Alice', 'Bob', 'Joe', 'Lester'],
                                           size=len(data)),
                          np.random.rand(len(data)) * 100,
                          np.random.randint(100, size=3),
                          rng))


cities_data = [['Alice', 'NYC'],
               ['Bob', 'Boston']]

df = pd.DataFrame(data, columns=['name', 'amount', 'id'])
date_df = pd.DataFrame(date_data, columns=['name', 'amount', 'id', 'ds'])
cities_df = pd.DataFrame(cities_data, columns=['name', 'city'])


# sc is from conftest.py


@pytest.yield_fixture(scope='module')
def sql(sc):
    try:
        yield HiveContext(sc)
    finally:
        dbpath = 'metastore_db'
        logpath = 'derby.log'
        if os.path.exists(dbpath):
            assert os.path.isdir(dbpath), '%s is not a directory' % dbpath
            shutil.rmtree(dbpath)
        if os.path.exists(logpath):
            assert os.path.isfile(logpath), '%s is not a file' % logpath
            os.remove(logpath)


@pytest.yield_fixture(scope='module')
def people(sc):
    with tmpfile('.txt') as fn:
        df.to_csv(fn, header=False, index=False)
        raw = sc.textFile(fn)
        parts = raw.map(lambda line: line.split(','))
        yield parts.map(
            lambda person: Row(
                name=person[0], amount=float(person[1]), id=int(person[2])
            )
        )


@pytest.yield_fixture(scope='module')
def cities(sc):
    with tmpfile('.txt') as fn:
        cities_df.to_csv(fn, header=False, index=False)
        raw = sc.textFile(fn)
        parts = raw.map(lambda line: line.split(','))
        yield parts.map(lambda person: Row(name=person[0], city=person[1]))


@pytest.yield_fixture(scope='module')
def date_people(sc):
    with tmpfile('.txt') as fn:
        date_df.to_csv(fn, header=False, index=False)
        raw = sc.textFile(fn)
        parts = raw.map(lambda line: line.split(','))
        yield parts.map(
            lambda person: Row(
                name=person[0],
                amount=float(person[1]),
                id=int(person[2]),
                ds=pd.Timestamp(person[3]).to_pydatetime()
            )
        )


@pytest.fixture(scope='module')
def ctx(sql, people, cities, date_people):
    sql.registerDataFrameAsTable(sql.createDataFrame(people), 't')
    sql.cacheTable('t')
    sql.registerDataFrameAsTable(sql.createDataFrame(cities), 's')
    sql.cacheTable('s')
    sql.registerDataFrameAsTable(sql.createDataFrame(date_people), 'dates')
    sql.cacheTable('dates')
    return sql


@pytest.fixture(scope='module')
def db(ctx):
    return symbol('db', discover(ctx))


def test_projection(db, ctx):
    expr = db.t[['id', 'name']]
    result = compute(expr, ctx, return_type='native')
    expected = compute(expr, {db: {'t': df}}, return_type='native')
    assert into(set, result) == into(set, expected)


def test_symbol_compute(db, ctx):
    assert isinstance(compute(db.t, ctx, return_type='native'), SparkDataFrame)


def test_field_access(db, ctx):
    for field in db.t.fields:
        expr = getattr(db.t, field)
        result = into(pd.Series, compute(expr, ctx, return_type='native'))
        expected = compute(expr, {db: {'t': df}}, return_type='native')
        assert result.name == expected.name
        np.testing.assert_array_equal(result.values,
                                      expected.values)


def test_head(db, ctx):
    expr = db.t[['name', 'amount']].head(2)
    result = compute(expr, ctx, return_type='native')
    expected = compute(expr, {db: {'t': df}}, return_type='native')
    assert into(list, result) == into(list, expected)


def test_literals(db, ctx):
    expr = db.t[db.t.amount >= 100]
    result = compute(expr, ctx, return_type='native')
    expected = compute(expr, {db: {'t': df}}, return_type='native')
    assert list(map(set, into(list, result))) == list(
        map(set, into(list, expected))
    )


def test_by_summary(db, ctx):
    t = db.t
    expr = by(t.name, mymin=t.amount.min(), mymax=t.amount.max())
    result = compute(expr, ctx, return_type='native')
    expected = compute(expr, {db: {'t': df}}, return_type='native')
    assert into(set, result) == into(set, expected)


def test_join(db, ctx):
    expr = join(db.t, db.s)
    result = compute(expr, ctx, return_type='native')
    expected = compute(expr, {db: {'t': df, 's': cities_df}}, return_type='native')

    assert isinstance(result, SparkDataFrame)
    assert into(set, result) == into(set, expected)
    assert discover(result) == expr.dshape


def test_join_diff_contexts(db, ctx, cities):
    expr = join(db.t, db.s, 'name')
    people = ctx.table('t')
    cities = into(ctx, cities, dshape=discover(ctx.table('s')))
    scope = {db: {'t': people, 's': cities}}
    result = compute(expr, scope, return_type='native')
    expected = compute(expr, {db: {'t': df, 's': cities_df}}, return_type='native')
    assert set(map(frozenset, odo(result, set))) == set(
        map(frozenset, odo(expected, set))
    )


def test_field_distinct(ctx, db):
    expr = db.t.name.distinct()
    result = compute(expr, ctx, return_type='native')
    expected = compute(expr, {db: {'t': df}}, return_type='native')
    assert into(set, result, dshape=expr.dshape) == into(set, expected)


def test_boolean(ctx, db):
    expr = db.t.amount > 50
    result = compute(expr, ctx, return_type='native')
    expected = compute(expr, {db: {'t': df}}, return_type='native')
    assert into(set, result, dshape=expr.dshape) == into(set, expected)


def test_selection(ctx, db):
    expr = db.t[db.t.amount > 50]
    result = compute(expr, ctx, return_type='native')
    expected = compute(expr, {db: {'t': df}}, return_type='native')
    assert list(map(set, into(list, result))) == list(
        map(set, into(list, expected))
    )


def test_selection_field(ctx, db):
    expr = db.t[db.t.amount > 50].name
    result = compute(expr, ctx, return_type='native')
    expected = compute(expr, {db: {'t': df}}, return_type='native')
    assert into(set, result, dshape=expr.dshape) == into(set, expected)


@pytest.mark.parametrize(
    ['field', 'reduction'],
    itertools.product(
        ['id', 'amount'],
        ['sum', 'max', 'min', 'mean', 'count', 'nunique']
    )
)
def test_reductions(ctx, db, field, reduction):
    expr = getattr(db.t[field], reduction)()
    result = compute(expr, ctx, return_type='native')
    expected = compute(expr, {db: {'t': df}}, return_type='native')
    assert into(list, result)[0][0] == expected


def test_column_arithmetic(ctx, db):
    expr = db.t.amount + 1
    result = compute(expr, ctx, return_type='native')
    expected = compute(expr, {db: {'t': df}}, return_type='native')
    assert into(set, result, dshape=expr.dshape) == into(set, expected)


@pytest.mark.parametrize('func', [sin, cos, tan, exp])
def test_math(ctx, db, func):
    expr = func(db.t.amount)
    result = compute(expr, ctx, return_type='native')
    expected = compute(expr, {db: {'t': df}}, return_type='native')
    np.testing.assert_allclose(
        np.sort(odo(result, np.ndarray, dshape=expr.dshape)),
        np.sort(odo(expected, np.ndarray))
    )


@pytest.mark.parametrize(['field', 'ascending'],
                         itertools.product(['name', 'id', ['name', 'amount']],
                                           [True, False]))
def test_sort(ctx, db, field, ascending):
    expr = db.t.sort(field, ascending=ascending)
    result = compute(expr, ctx, return_type='native')
    expected = compute(expr, {db: {'t': df}}, return_type='native')
    assert list(map(set, into(list, result))) == list(
        map(set, into(list, expected))
    )


@pytest.mark.xfail
def test_map(ctx, db):
    expr = db.t.id.map(lambda x: x + 1, 'int')
    result = compute(expr, ctx, return_type='native')
    expected = compute(expr, {db: {'t': df}}, return_type='native')
    assert into(set, result, dshape=expr.dshape) == into(set, expected)


@pytest.mark.parametrize(
    ['grouper', 'reducer', 'reduction'],
    itertools.chain(
        itertools.product(
            ['name', 'id', ['id', 'amount']],
            ['id', 'amount'],
            ['sum', 'count', 'max', 'min', 'mean', 'nunique']
        ),
        [('name', 'name', 'count'), ('name', 'name', 'nunique')]
    )
)
def test_by(ctx, db, grouper, reducer, reduction):
    t = db.t
    expr = by(t[grouper], total=getattr(t[reducer], reduction)())
    result = compute(expr, ctx, return_type='native')
    expected = compute(expr, {db: {'t': df}}, return_type='native')
    assert set(map(frozenset, into(list, result))) == set(
        map(frozenset, into(list, expected))
    )


@pytest.mark.parametrize(
    ['reducer', 'reduction'],
    itertools.product(['id', 'name'], ['count', 'nunique'])
)
def test_multikey_by(ctx, db, reducer, reduction):
    t = db.t
    expr = by(t[['id', 'amount']], total=getattr(t[reducer], reduction)())
    result = compute(expr, ctx, return_type='native')
    expected = compute(expr, {db: {'t': df}}, return_type='native')
    assert (set(map(frozenset, into(list, result))) ==
            set(map(frozenset, into(list, expected))))


def test_grouper_with_arith(ctx, db):
    expr = by(db.t[['id', 'amount']], total=(db.t.amount + 1).sum())
    result = compute(expr, ctx, return_type='native')
    expected = compute(expr, {db: {'t': df}}, return_type='native')
    assert list(map(set, into(list, result))) == list(map(set, into(list, expected)))


def test_by_non_native_ops(ctx, db):
    expr = by(db.t.id, total=db.t.id.nunique())
    result = compute(expr, ctx, return_type='native')
    expected = compute(expr, {db: {'t': df}}, return_type='native')
    assert list(map(set, into(list, result))) == list(map(set, into(list, expected)))


def test_str_len(ctx, db):
    expr = db.t.name.str.len()
    result = odo(compute(expr, ctx, return_type='native'), pd.Series)
    expected = compute(expr, {db: {'t': df}}, return_type='native')
    assert result.name == 'name'
    assert expected.name == 'name'
    assert odo(result, set) == odo(expected, set)


@pytest.mark.parametrize(
    'attr',
    ['year', 'month', 'day', 'hour', 'minute', 'second'] + list(
        map(
            partial(
                pytest.mark.xfail,
                raises=(Py4JJavaError, AnalysisException)
            ),
            ['millisecond', 'microsecond']
        )
    )
)
def test_by_with_date(ctx, db, attr):
    # TODO: investigate CSV writing precision between pandas 0.16.0 and 0.16.1
    # TODO: see if we can use odo to convert the dshape of an existing
    #       DataFrame
    expr = by(getattr(db.dates.ds, attr), mean=db.dates.amount.mean())
    result = odo(
        compute(expr, ctx, return_type='native'), pd.DataFrame
    ).sort('mean').reset_index(drop=True)
    expected = compute(
        expr,
        {db: {'dates': date_df}},
        return_type='native'
    ).sort('mean').reset_index(drop=True)
    tm.assert_frame_equal(result, expected, check_dtype=False)


@pytest.mark.parametrize('keys', [[1], [1, 2]])
def test_isin(ctx, db, keys):
    expr = db.t[db.t.id.isin(keys)]
    result = odo(compute(expr, ctx, return_type='native'), set)
    expected = odo(compute(expr, {db: {'t': df}}, return_type='native'), set)
    assert (set(map(frozenset, odo(result, list))) ==
            set(map(frozenset, odo(expected, list))))


def test_nunique_spark_dataframe(ctx, db):
    result = odo(compute(db.t.nunique(), ctx, return_type='native'), int)
    expected = ctx.table('t').distinct().count()
    assert result == expected


def test_core_compute(ctx, db):
    assert isinstance(compute(db.t, ctx, return_type='core'), pd.DataFrame)
    assert isinstance(compute(db.t.amount, ctx, return_type='core'), pd.Series)
    assert iscorescalar(compute(db.t.amount.mean(), ctx, return_type='core'))
    assert isinstance(compute(db.t, ctx, return_type=list), list)



from __future__ import absolute_import, division, print_function

import operator
from toolz import first
import numpy as np
import pandas as pd
from datashape import (
    DataShape,
    DateTime,
    Option,
    String,
    TimeDelta,
    coretypes as ct,
    datetime_,
    discover,
    dshape,
    optionify,
    promote,
    timedelta_,
    unsigned,
)
from datashape.predicates import isscalar, isboolean, isnumeric, isdatelike
from datashape.typesets import integral
from dateutil.parser import parse as dt_parse


from .core import parenthesize, eval_str
from .expressions import Expr, shape, ElemWise, binop_inputs, binop_name
from .utils import maxshape
from ..dispatch import dispatch
from ..compatibility import _strtypes


__all__ = '''
BinOp
UnaryOp
Arithmetic
Add
Mult
Repeat
Sub
Div
FloorDiv
Pow
Mod
Interp
USub
Relational
Eq
Ne
Ge
Lt
Le
Gt
Gt
And
Or
Not
'''.split()


class BinOp(ElemWise):
    _arguments = 'lhs', 'rhs'
    __inputs__ = 'lhs', 'rhs'

    def __str__(self):
        lhs = parenthesize(eval_str(self.lhs))
        rhs = parenthesize(eval_str(self.rhs))
        return '%s %s %s' % (lhs, self.symbol, rhs)

    def _dshape(self):
        # TODO: better inference.  e.g. int + int -> int
        return DataShape(*(maxshape([shape(self.lhs), shape(self.rhs)]) +
                           (self._dtype,)))

    _name = property(binop_name)


    @property
    def _inputs(self):
        return tuple(binop_inputs(self))


class UnaryOp(ElemWise):
    _arguments = '_child',

    def __str__(self):
        return '%s(%s)' % (self.symbol, eval_str(self._child))

    @property
    def symbol(self):
        return type(self).__name__

    def _dshape(self):
        return DataShape(*(shape(self._child) + (self._dtype,)))

    @property
    def _name(self):
        return self._child._name


class Arithmetic(BinOp):
    """ Super class for arithmetic operators like add or mul """

    @property
    def _dtype(self):
        # we can't simply use .schema or .datashape because we may have a bare
        # integer, for example
        lhs, rhs = discover(self.lhs).measure, discover(self.rhs).measure
        return promote(lhs, rhs)


class Add(Arithmetic):
    symbol = '+'
    op = operator.add

    @property
    def _dtype(self):
        lmeasure = discover(self.lhs).measure
        lty = getattr(lmeasure, 'ty', lmeasure)
        rmeasure = discover(self.rhs).measure
        rty = getattr(rmeasure, 'ty', rmeasure)
        if lmeasure == datetime_ and rmeasure == datetime_:
            raise TypeError('cannot add datetime to datetime')

        l_is_datetime = lty == datetime_
        if l_is_datetime or rty == datetime_:
            if l_is_datetime:
                other = rty
            else:
                other = lty
            if isinstance(other, TimeDelta):
                return optionify(lmeasure, rmeasure, datetime_)
            else:
                raise TypeError(
                    'can only add timedeltas to datetimes',
                )

        return super(Add, self)._dtype


class Mult(Arithmetic):
    symbol = '*'
    op = operator.mul


class Repeat(Arithmetic):
    # Sequence repeat
    symbol = '*'
    op = operator.mul

    @property
    def _dtype(self):
        lmeasure = discover(self.lhs).measure
        rmeasure = discover(self.rhs).measure
        if not (isinstance(getattr(lmeasure, 'ty', lmeasure), String) and
                getattr(rmeasure, 'ty', rmeasure) in integral):
            raise TypeError(
                'can only repeat strings by an integer amount, got: %s * %s' %
                (lmeasure, rmeasure),
            )

        return optionify(lmeasure, rmeasure, lmeasure)


class Sub(Arithmetic):
    symbol = '-'
    op = operator.sub

    @property
    def _dtype(self):
        lmeasure = discover(self.lhs).measure
        lty = getattr(lmeasure, 'ty', lmeasure)
        rmeasure = discover(self.rhs).measure
        rty = getattr(rmeasure, 'ty', rmeasure)
        if lty == datetime_:
            if isinstance(rty, DateTime):
                return optionify(lmeasure, rmeasure, timedelta_)
            if isinstance(rty, TimeDelta):
                return optionify(lmeasure, rmeasure, datetime_)
            else:
                raise TypeError(
                    'can only subtract timedelta or datetime from datetime',
                )

        return super(Sub, self)._dtype


class Div(Arithmetic):
    symbol = '/'
    op = operator.truediv

    @property
    def _dtype(self):
        lhs, rhs = discover(self.lhs).measure, discover(self.rhs).measure
        return optionify(lhs, rhs, ct.float64)


class FloorDiv(Arithmetic):
    symbol = '//'
    op = operator.floordiv

    @property
    def _dtype(self):
        lhs, rhs = discover(self.lhs).measure, discover(self.rhs).measure
        is_unsigned = lhs in unsigned and rhs in unsigned
        max_width = max(lhs.itemsize, rhs.itemsize)
        prefix = 'u' if is_unsigned else ''
        measure = getattr(ct, '%sint%d' % (prefix, max_width * 8))
        return optionify(lhs, rhs, measure)


class Pow(Arithmetic):
    symbol = '**'
    op = operator.pow


class Mod(Arithmetic):
    symbol = '%'
    op = operator.mod


class Interp(Arithmetic):
    # String interpolation
    symbol = '%'
    op = operator.mod

    @property
    def _dtype(self):
        lmeasure = discover(self.lhs).measure
        rmeasure = discover(self.rhs).measure
        if not (isinstance(getattr(lmeasure, 'ty', lmeasure), String)):
            raise TypeError('can only interp strings got: %s' % lmeasure)

        return optionify(lmeasure, rmeasure, lmeasure)


class USub(UnaryOp):
    op = operator.neg
    symbol = '-'

    def __str__(self):
        return '-%s' % parenthesize(eval_str(self._child))

    @property
    def _dtype(self):
        # TODO: better inference.  -uint -> int
        return self._child.schema


@dispatch(ct.Option, object)
def scalar_coerce(ds, val):
    return scalar_coerce(ds.ty, val) if val is not None else None


@dispatch((ct.Record, ct.Mono, ct.Option, DataShape), Expr)
def scalar_coerce(ds, val):
    return val


@dispatch(ct.Date, _strtypes)
def scalar_coerce(_, val):
    if val == '':
        raise TypeError('%r is not a valid date' % val)
    dt = dt_parse(val)
    if any(x > 0 for x in (dt.hour, dt.minute, dt.second, dt.microsecond)):
        msg = "Can not coerce %r to type Date, contains time information" 
        raise TypeError(msg % val)
    return dt.date()


@dispatch(ct.DateTime, _strtypes)
def scalar_coerce(_, val):
    if val == '':
        raise TypeError('%r is not a valid datetime' % val)
    return pd.Timestamp(val)


@dispatch(ct.CType, _strtypes)
def scalar_coerce(dt, val):
    return np.asscalar(np.asarray(val, dtype=dt.to_numpy_dtype()))


@dispatch(ct.Record, object)
def scalar_coerce(rec, val):
    if len(rec.fields) == 1:
        return scalar_coerce(first(rec.types), val)
    else:
        raise TypeError("Trying to coerce complex datashape\n"
                "got dshape: %s\n"
                "scalar_coerce only intended for scalar values" % rec)


@dispatch(ct.DataShape, object)
def scalar_coerce(ds, val):
    return scalar_coerce(ds.measure, val)


@dispatch(object, object)
def scalar_coerce(dtype, val):
    return val


@dispatch(_strtypes, object)
def scalar_coerce(ds, val):
    return scalar_coerce(dshape(ds), val)


def _neg(self):
    return USub(self)


def _mkbin(name, cons, private=True, reflected=True):
    prefix = '_' if private else ''

    def _bin(self, other):
        result = cons(self, scalar_coerce(self.dshape, other))
        result.dshape  # Check that shapes and dtypes match up
        return result
    _bin.__name__ = prefix + name

    if reflected:
        def _rbin(self, other):
            result = cons(scalar_coerce(self.dshape, other), self)
            result.dshape  # Check that shapes and dtypes match up
            return result
        _rbin.__name__ = prefix + 'r' + name

        return _bin, _rbin

    return _bin


_add, _radd = _mkbin('add', Add)
_div, _rdiv = _mkbin('div', Div)
_floordiv, _rfloordiv = _mkbin('floordiv', FloorDiv)
_mod, _rmod = _mkbin('mod', Mod)
_mul, _rmul = _mkbin('mul', Mult)
_pow, _rpow = _mkbin('pow', Pow)
repeat = _mkbin('repeat', Repeat, reflected=False, private=False)
_sub, _rsub = _mkbin('sub', Sub)
interp = _mkbin('interp', Interp, reflected=False, private=False)


class _Optional(Arithmetic):
    @property
    def _dtype(self):
        # we can't simply use .schema or .datashape because we may have a bare
        # integer, for example
        lhs, rhs = discover(self.lhs).measure, discover(self.rhs).measure
        if isinstance(lhs, Option) or isinstance(rhs, Option):
            return Option(ct.bool_)
        return ct.bool_


class Relational(_Optional):
    # Leave this to separate relationals from other types of optionals.
    pass


class Eq(Relational):
    symbol = '=='
    op = operator.eq


class Ne(Relational):
    symbol = '!='
    op = operator.ne


class Ge(Relational):
    symbol = '>='
    op = operator.ge


class Le(Relational):
    symbol = '<='
    op = operator.le


class Gt(Relational):
    symbol = '>'
    op = operator.gt


class Lt(Relational):
    symbol = '<'
    op = operator.lt


class And(_Optional):
    symbol = '&'
    op = operator.and_


class Or(_Optional):
    symbol = '|'
    op = operator.or_


class Not(UnaryOp):
    symbol = '~'
    op = operator.invert

    @property
    def _dtype(self):
        return self._child.schema

    def __str__(self):
        return '~%s' % parenthesize(eval_str(self._child))


_and, _rand = _mkbin('and', And)
_eq = _mkbin('eq', Eq, reflected=False)
_ge = _mkbin('ge', Ge, reflected=False)
_gt = _mkbin('gt', Gt, reflected=False)
_le = _mkbin('le', Le, reflected=False)
_lt = _mkbin('lt', Lt, reflected=False)
_ne = _mkbin('ne', Ne, reflected=False)
_or, _ror = _mkbin('or', Or)


def _invert(self):
    result = Invert(self)
    result.dshape  # Check that shapes and dtypes match up
    return result


Invert = Not
BitAnd = And
BitOr = Or


from .expressions import schema_method_list


schema_method_list.extend([
    (isnumeric,
     set([_add, _radd, _mul, _rmul, _div, _rdiv, _floordiv, _rfloordiv, _sub,
          _rsub, _pow, _rpow, _mod, _rmod,  _neg])),
    (isscalar, set([_eq, _ne, _lt, _le, _gt, _ge])),
    (isboolean, set([_or, _ror, _and, _rand, _invert])),
    (isdatelike, set([_add, _radd, _sub, _rsub])),
    ])



from __future__ import absolute_import, division, print_function

import datashape
from datashape import Record, DataShape, dshape, TimeDelta, Decimal, Option
from datashape import coretypes as ct
from datashape.predicates import iscollection, isboolean, isnumeric, isdatelike
from numpy import inf
from odo.utils import copydoc
import toolz

from .core import common_subexpression
from .expressions import Expr, ndim, dshape_method_list, method_properties
from .strings import isstring


def _normalize_axis(axis, child):
    if axis is None:
        axis = tuple(range(child.ndim))
    if isinstance(axis, (set, list)):
        axis = tuple(axis)
    if not isinstance(axis, tuple):
        axis = (axis,)
    return tuple(sorted(axis))


class Reduction(Expr):

    """ A column-wise reduction

    Blaze supports the same class of reductions as NumPy and Pandas.

        sum, min, max, any, all, mean, var, std, count, nunique

    Examples
    --------

    >>> from blaze import symbol
    >>> t = symbol('t', 'var * {name: string, amount: int, id: int}')
    >>> e = t['amount'].sum()

    >>> data = [['Alice', 100, 1],
    ...         ['Bob', 200, 2],
    ...         ['Alice', 50, 3]]

    >>> from blaze.compute.python import compute
    >>> compute(e, data)
    350
    """
    _arguments = '_child', 'axis', 'keepdims'

    def __new__(cls, _child, axis=None, keepdims=False):
        axis = _normalize_axis(axis, _child)
        return super(Reduction, cls).__new__(cls, _child, axis, keepdims)

    def _dshape(self):
        axis = self.axis
        if self.keepdims:
            shape = tuple(1 if i in axis else d
                          for i, d in enumerate(self._child.shape))
        else:
            shape = tuple(d
                          for i, d in enumerate(self._child.shape)
                          if i not in axis)
        return DataShape(*(shape + (self.schema,)))

    def _schema(self):
        schema = self._child.schema[0]
        if isinstance(schema, Record) and len(schema.types) == 1:
            result = toolz.first(schema.types)
        else:
            result = schema
        return DataShape(result)

    @property
    def symbol(self):
        return type(self).__name__

    @property
    def _name(self):
        child_name = self._child._name
        if child_name is None or child_name == '_':
            return type(self).__name__
        else:
            return '%s_%s' % (child_name, type(self).__name__)

    def __str__(self):
        kwargs = list()
        if self.keepdims:
            kwargs.append('keepdims=True')
        if self.axis != tuple(range(self._child.ndim)):
            kwargs.append('axis=' + str(self.axis))
        other = sorted(
            set(self._arguments[1:]) - set(['_child', 'axis', 'keepdims']))
        for slot in other:
            kwargs.append('%s=%s' % (slot, getattr(self, slot)))
        name = type(self).__name__
        if kwargs:
            return '%s(%s, %s)' % (name, self._child, ', '.join(kwargs))
        else:
            return '%s(%s)' % (name, self._child)


class any(Reduction):
    schema = dshape(ct.bool_)


class all(Reduction):
    schema = dshape(ct.bool_)


class sum(Reduction):
    def _schema(self):
        return DataShape(datashape.maxtype(super(sum, self)._schema()))


class max(Reduction):
    pass


class min(Reduction):
    pass


class FloatingReduction(Reduction):
    def _schema(self):
        measure = self._child.schema.measure
        base = getattr(measure, 'ty', measure)
        return_type = Option if isinstance(measure, Option) else toolz.identity
        return DataShape(return_type(
            base
            if isinstance(base, Decimal) else
            base
            if isinstance(base, TimeDelta) else
            ct.float64,
        ))


class mean(FloatingReduction):
    pass


class var(FloatingReduction):

    """Variance

    Parameters
    ----------
    child : Expr
        An expression
    unbiased : bool, optional
        Compute an unbiased estimate of the population variance if this is
        ``True``. In NumPy and pandas, this parameter is called ``ddof`` (delta
        degrees of freedom) and is equal to 1 for unbiased and 0 for biased.
    """
    _arguments = '_child', 'unbiased', 'axis', 'keepdims'

    def __new__(cls, child, unbiased=False, axis=None, keepdims=False):
        axis = _normalize_axis(axis, child)
        return super(Reduction, cls).__new__(
            cls, child, unbiased, axis, keepdims,
        )


class std(FloatingReduction):

    """Standard Deviation

    Parameters
    ----------
    child : Expr
        An expression
    unbiased : bool, optional
        Compute the square root of an unbiased estimate of the population
        variance if this is ``True``.

        .. warning::

            This does *not* return an unbiased estimate of the population
            standard deviation.

    See Also
    --------
    var
    """
    _arguments = '_child', 'unbiased', 'axis', 'keepdims'

    def __new__(cls, child, unbiased=False, axis=None, keepdims=False):
        axis = _normalize_axis(axis, child)
        return super(Reduction, cls).__new__(
            cls, child, unbiased, axis, keepdims,
        )


class count(Reduction):

    """ The number of non-null elements """
    schema = dshape(ct.int32)


class nunique(Reduction):
    schema = dshape(ct.int32)


class nelements(Reduction):

    """Compute the number of elements in a collection, including missing values.

    See Also
    ---------
    blaze.expr.reductions.count: compute the number of non-null elements

    Examples
    --------
    >>> from blaze import symbol
    >>> t = symbol('t', 'var * {name: string, amount: float64}')
    >>> t[t.amount < 1].nelements()
    nelements(t[t.amount < 1])
    """
    schema = dshape(ct.int32)


def nrows(expr):
    return nelements(expr, axis=(0,))


class Summary(Expr):

    """ A collection of named reductions

    Examples
    --------

    >>> from blaze import symbol
    >>> t = symbol('t', 'var * {name: string, amount: int, id: int}')
    >>> expr = summary(number=t.id.nunique(), sum=t.amount.sum())

    >>> data = [['Alice', 100, 1],
    ...         ['Bob', 200, 2],
    ...         ['Alice', 50, 1]]

    >>> from blaze import compute
    >>> compute(expr, data)
    (2, 350)
    """
    _arguments = '_child', 'names', 'values', 'axis', 'keepdims'

    def __new__(cls, _child, names, values, axis=None, keepdims=False):
        return super(Summary, cls).__new__(
            cls, _child, names, values, axis, keepdims,
        )

    def _dshape(self):
        axis = self.axis
        if self.keepdims:
            shape = tuple(1 if i in axis else d
                          for i, d in enumerate(self._child.shape))
        else:
            shape = tuple(d
                          for i, d in enumerate(self._child.shape)
                          if i not in axis)
        measure = Record(list(zip(self.names,
                                  [v.schema for v in self.values])))
        return DataShape(*(shape + (measure,)))

    def __str__(self):
        s = 'summary('
        s += ', '.join('%s=%s' % (name, str(val))
                       for name, val in zip(self.fields, self.values))
        if self.keepdims:
            s += ', keepdims=True'
        s += ')'
        return s


@copydoc(Summary)
def summary(keepdims=False, axis=None, **kwargs):
    items = sorted(kwargs.items(), key=toolz.first)
    names = tuple(map(toolz.first, items))
    values = tuple(map(toolz.second, items))
    child = common_subexpression(*values)

    if len(kwargs) == 1 and not iscollection(child.dshape):
        while not iscollection(child.dshape):
            children = [i for i in child._inputs if isinstance(i, Expr)]
            if len(children) == 1:
                child = children[0]
            else:
                child = common_subexpression(*children)

    if axis is None:
        axis = tuple(range(ndim(child)))
    if isinstance(axis, (set, list)):
        axis = tuple(axis)
    if not isinstance(axis, tuple):
        axis = (axis,)
    return Summary(child, names, values, keepdims=keepdims, axis=axis)


def vnorm(expr, ord=None, axis=None, keepdims=False):
    """ Vector norm

    See np.linalg.norm
    """
    if ord is None or ord == 'fro':
        ord = 2
    if ord == inf:
        return max(abs(expr), axis=axis, keepdims=keepdims)
    elif ord == -inf:
        return min(abs(expr), axis=axis, keepdims=keepdims)
    elif ord == 1:
        return sum(abs(expr), axis=axis, keepdims=keepdims)
    elif ord % 2 == 0:
        return sum(expr ** ord, axis=axis, keepdims=keepdims) ** (1.0 / ord)
    return sum(abs(expr) ** ord, axis=axis, keepdims=keepdims) ** (1.0 / ord)


dshape_method_list.extend([
    (iscollection, set([count, nelements])),
    (lambda ds: (iscollection(ds) and
                 (isstring(ds) or isnumeric(ds) or isboolean(ds) or
                  isdatelike(ds))),
     set([min, max])),
    (lambda ds: len(ds.shape) == 1,
     set([nrows, nunique])),
    (lambda ds: iscollection(ds) and isboolean(ds),
     set([any, all])),
    (lambda ds: iscollection(ds) and (isnumeric(ds) or isboolean(ds)),
     set([mean, sum, std, var, vnorm])),
])

method_properties.update([nrows])



from blaze.expr import symbol
import numpy as np
from datashape import dshape, isscalar


def test_array_dshape():
    x = symbol('x', '5 * 3 * float32')
    assert x.shape == (5, 3)
    assert x.schema == dshape('float32')
    assert len(x) == 5
    assert x.ndim == 2


def test_element():
    x = symbol('x', '5 * 3 * float32')
    assert isscalar(x[1, 2].dshape)
    assert x[1, 2].dshape == dshape('float32')

    assert str(x[1, 2]) == 'x[1, 2]'

    x = symbol('x', '5 * float32')
    assert isscalar(x[3].dshape)


def test_slice():
    x = symbol('x', '5 * 3 * {name: string, amount: float32}')
    assert x[2:, 0].dshape == dshape('3 * {name: string, amount: float32}')

    assert x[2:].dshape == x[2:, :].dshape

    # Make sure that these are hashable
    hash(x[:2])
    hash(x[0, :2])

    assert str(x[1]) == 'x[1]'
    assert str(x[:2]) == 'x[:2]'
    assert str(x[0, :2]) == 'x[0, :2]'
    assert str(x[1:4:2, :2]) == 'x[1:4:2, :2]'


def test_negative_slice():
    x = symbol('x', '10 * 10 * int32')
    assert x[:5, -3:].shape == (5, 3)


def test_None_slice():
    x = symbol('x', '10 * 10 * int32')
    assert x[:5, None, -3:].shape == (5, 1, 3)


def test_list_slice():
    x = symbol('x', '10 * 10 * int32')
    assert x[[1, 2, 3], [4, 5]].shape == (3, 2)


def test_list_slice_string():
    x = symbol('x', '10 * 10 * int32')
    assert str(x[[1, 2, 3]]) == "x[[1, 2, 3]]"


def test_slice_with_boolean_list():
    x = symbol('x', '5 * int32')
    expr = x[[True, False, False, True, False]]
    assert expr.index == ([0, 3],)


def test_slice_with_numpy_array():
    x = symbol('x', '2 * int32')
    assert x[np.array([True, False])].isidentical(x[[True, False]])
